# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import sys

import pytest

from pex.pep_440 import Version
from pex.specifier_sets import ExcludedRange, LowerBound, Range, UpperBound, as_range, includes
from pex.third_party.packaging.specifiers import InvalidSpecifier
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    pass


def test_range_simplification():
    # type: () -> None

    # No simplification.
    assert (
        Range(
            lower=LowerBound(Version("2.7"), inclusive=True),
            upper=UpperBound(Version("3.13"), inclusive=False),
            excludes=tuple(
                [
                    ExcludedRange(
                        lower=LowerBound(Version("3"), inclusive=True),
                        upper=UpperBound(Version("3.1"), inclusive=False),
                    ),
                    ExcludedRange(
                        lower=LowerBound(Version("3.1"), inclusive=True),
                        upper=UpperBound(Version("3.2"), inclusive=False),
                    ),
                    ExcludedRange(
                        lower=LowerBound(Version("3.2"), inclusive=True),
                        upper=UpperBound(Version("3.3"), inclusive=False),
                    ),
                    ExcludedRange(
                        lower=LowerBound(Version("3.3"), inclusive=True),
                        upper=UpperBound(Version("3.4"), inclusive=False),
                    ),
                    ExcludedRange(
                        lower=LowerBound(Version("3.4"), inclusive=True),
                        upper=UpperBound(Version("3.5"), inclusive=False),
                    ),
                ]
            ),
        )
    ) == as_range(">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*,<3.13")

    # Bounds narrowing and exclude trimming.
    assert (
        Range(
            lower=LowerBound(Version("3"), inclusive=True),
            upper=UpperBound(Version("6"), inclusive=False),
            excludes=tuple(
                [
                    ExcludedRange(
                        lower=LowerBound(Version("4"), inclusive=True),
                        upper=UpperBound(Version("5"), inclusive=False),
                    )
                ]
            ),
        )
        == as_range(">0,!=1.*,>=2,!=2.*,!=4.*,<6,!=7.*,<42")
    )


def test_includes():
    # type: () -> None

    assert includes("", "")
    assert includes("", "==1")
    assert includes("", ">1")
    assert includes("", ">=1")
    assert includes("", "<1")
    assert includes("", "<=1")
    assert includes("", "==1.*")
    assert includes("", "!=1.*")
    assert includes("", "~=1.2")

    assert not includes("", "===bob")

    # Arbitrary equality: versions treated as strings and must match exactly.
    assert includes("===bob", "===bob")
    assert not includes("===fred", "===bob")
    assert includes("===1", "===1")
    assert not includes("===1.0", "===1")
    assert not includes("===1", "===1.0")

    # Compatible X.Y: should be equivalent to >=X.Y,==X.* or, equivalently >=X.Y,<X+1.
    assert includes("~=1.2", ">=1.2,==1.*")
    assert includes("~=1.2", ">=1.2,<2")
    assert includes("~=1.2", ">=1.3,<2")
    assert includes("~=1.2", ">=1.2,<1.9")
    assert includes("~=1.2", "==1.2.*")
    assert includes("~=1.2", "==1.3.*")
    assert includes("~=1.2", "==1.3")

    assert not includes("~=1.2", "==1.1")
    assert not includes("~=1.2", "==2")
    assert not includes("~=1.2", "==1.*")
    assert not includes("~=1.2", "")

    assert includes("~=1.2.3", ">=1.2.3,==1.2.*")
    assert includes("~=1.2.3", ">=1.2.3,<1.3")
    assert includes("~=1.2.3", "==1.2.3")
    assert includes("~=1.2.3", "==1.2.13")

    assert not includes("~=1.2.3", "==1.3")
    assert not includes("~=1.2.3", "==1.2.2")

    # Equality: missing components should be 0-filled for comparison.
    assert includes("==1", "==1")
    assert includes("==1.0", "==1")
    assert includes("==1", "==1.0")

    assert not includes("==1", "==1.0.1")
    assert not includes("==1", "==0.9")
    assert not includes("==1", "==1.*")

    assert includes("!=2", "==1")
    assert includes("!=2", "==3")
    assert includes("!=2", "==2.1")
    assert includes("!=2", "<2")
    assert includes("!=2", ">2")
    assert includes("!=2", ">2,<2.1")

    assert not includes("!=2", "==2")
    assert not includes("!=2", "==2.0")
    assert not includes("!=2", "~=2.0")
    assert not includes("!=2", ">=1,<3")

    # N.B.: The LHS is the empty range, and we make the policy call that no range contains the
    # empty range.
    assert not includes("!=2", ">2,<=2")
    assert not includes("!=2", ">2,<2")
    assert not includes("!=2", ">=2,<2")
    # And vice-versa, which involves no policy call.
    assert not includes(">2,<2", "")
    assert not includes(">2,<2", "==2")

    assert not includes("!=1", "")

    assert includes(">1", ">1")
    assert includes(">1.0", ">1")
    assert includes(">1", ">1.0")

    # Newer versions of vendored packaging used for Python>=3.7 enforce the .* suffix is only used
    # with == and != operators; so we don't test for those versions.
    if sys.version_info[:2] < (3, 7):
        assert includes(">1.*", ">1")
        assert includes(">1.*", ">1.0")
        assert includes(">1", ">1.*")
        assert includes(">1.0", ">1.*")
        assert includes(">1.*", ">1.*")
    else:
        with pytest.raises(InvalidSpecifier):
            as_range(">1.*")

    complex_requirement = ">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*,<3.13"
    assert includes(complex_requirement, "==2.7")
    assert includes(complex_requirement, "==2.7.18")
    assert includes(complex_requirement, "==3.5")
    assert includes(complex_requirement, "==3.12.3")

    assert not includes(complex_requirement, "==2.6")
    assert not includes(complex_requirement, "==3")
    assert not includes(complex_requirement, "==3.0")
    assert not includes(complex_requirement, "==3.0.0")
    assert not includes(complex_requirement, "==3.0.1")
    assert not includes(complex_requirement, "==3.1")
    assert not includes(complex_requirement, "==3.2")
    assert not includes(complex_requirement, "==3.3")
    assert not includes(complex_requirement, "==3.4")
    assert not includes(complex_requirement, "==3.4.99")
    assert not includes(complex_requirement, "==3.13")
    assert not includes(complex_requirement, ">=2.6,<3.14")

    assert includes("", "!=2.*")
    assert not includes("==2", "!=2.*")
    assert not includes(">2", "!=2.*")
    assert not includes("<=2", "!=2.*")

    assert not includes(">=1,!=2,<4", ">=1,<4")
    assert includes(">=1,<4", ">=1,!=2,<4")
    assert not includes(">=1,!=2,!=3,<4", ">=1,!=2,<4")
    assert includes("!=2", ">=1,!=2,<4")

    assert includes(">=2,!=3.*,<4", ">=1,!=1.*,<3")
    assert includes(">=2,!=3.*,<4", ">=1,!=1.*,<3,>0,<6")
    assert includes(">=2,!=3.*,<4,>=1,<=42", ">=1,!=1.*,<3")


def test_compatible_release_version_suffix_handling():
    # type: () -> None

    # Dev-release specifiers.
    assert (
        Range(
            lower=LowerBound(Version("1.4.5.dev4"), inclusive=True),
            upper=UpperBound(Version("1.5"), inclusive=False),
        )
        == as_range("~=1.4.5.dev4")
    )

    # Pre-release specifiers.
    assert (
        Range(
            lower=LowerBound(Version("1.4.5a4"), inclusive=True),
            upper=UpperBound(Version("1.5"), inclusive=False),
        )
        == as_range("~=1.4.5a4")
    )
    assert (
        Range(
            lower=LowerBound(Version("1.4.5b4"), inclusive=True),
            upper=UpperBound(Version("1.5"), inclusive=False),
        )
        == as_range("~=1.4.5b4")
    )
    assert (
        Range(
            lower=LowerBound(Version("1.4.5rc4"), inclusive=True),
            upper=UpperBound(Version("1.5"), inclusive=False),
        )
        == as_range("~=1.4.5rc4")
    )

    # Post-release specifiers.
    assert (
        Range(
            lower=LowerBound(Version("2.2.post3"), inclusive=True),
            upper=UpperBound(Version("3"), inclusive=False),
        )
        == as_range("~=2.2.post3")
    )

    # Local-release specifiers.
    with pytest.raises(InvalidSpecifier):
        as_range("~=1.4.5+bob")


def assert_wildcard_handling(version_suffix):
    # type: (str) -> None

    version_to_wildcard = "1.4.5{version_suffix}".format(version_suffix=version_suffix)

    # N.B.: Newer versions of vendored packaging used for Python>=3.7 enforce the .* suffix is
    # only used with un-suffixed versions, so we test both old and new vendored packaging
    # expectations.

    eq_specifier = "=={version_to_wildcard}.*".format(version_to_wildcard=version_to_wildcard)
    if sys.version_info[:2] < (3, 7):
        assert (
            Range(
                lower=LowerBound(Version(version_to_wildcard), inclusive=True),
                upper=UpperBound(Version("1.4.6"), inclusive=False),
            )
            == as_range(eq_specifier)
        )
    else:
        with pytest.raises(InvalidSpecifier):
            as_range(eq_specifier)

    ne_specifier = "!={version_to_wildcard}.*".format(version_to_wildcard=version_to_wildcard)
    if sys.version_info[:2] < (3, 7):
        assert (
            Range(
                excludes=tuple(
                    [
                        ExcludedRange(
                            lower=LowerBound(Version(version_to_wildcard), inclusive=True),
                            upper=UpperBound(Version("1.4.6"), inclusive=False),
                        )
                    ]
                )
            )
            == as_range(ne_specifier)
        )
    else:
        with pytest.raises(InvalidSpecifier):
            as_range(ne_specifier)


def test_wildcard_version_suffix_handling():
    # type: () -> None

    # Dev-release specifiers.
    assert_wildcard_handling(".dev4")

    # Pre-release specifiers.
    assert_wildcard_handling("a4")
    assert_wildcard_handling("b4")
    assert_wildcard_handling("rc4")

    # Post-release specifiers.
    assert_wildcard_handling(".post3")

    # Local-release specifiers.
    assert_wildcard_handling("+bob")



# Copyright 2022 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import json
import os.path
import re
import sys
from textwrap import dedent

import pytest

from pex.dist_metadata import Requirement
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.pip.version import PipVersion
from pex.resolve.locked_resolve import (
    Artifact,
    LocalProjectArtifact,
    LockedRequirement,
    LockedResolve,
    LockStyle,
    VCSArtifact,
)
from pex.resolve.lockfile import json_codec
from pex.resolve.lockfile.model import Lockfile
from pex.resolve.resolved_requirement import ArtifactURL, Fingerprint, Pin
from pex.resolve.resolver_configuration import ResolverVersion
from pex.sorted_tuple import SortedTuple
from pex.typing import TYPE_CHECKING
from pex.version import __version__
from testing.cli import run_pex3

if TYPE_CHECKING:
    from typing import Any, Iterable, Optional, Text

    import attr  # vendor:skip
else:
    from pex.third_party import attr


UNIVERSAL_ANSICOLORS = Lockfile(
    pex_version="42",
    style=LockStyle.UNIVERSAL,
    requires_python=SortedTuple(),
    target_systems=SortedTuple(),
    pip_version=PipVersion.DEFAULT,
    resolver_version=ResolverVersion.PIP_2020,
    requirements=SortedTuple([Requirement.parse("ansicolors")]),
    constraints=SortedTuple(),
    allow_prereleases=False,
    allow_wheels=True,
    only_wheels=SortedTuple(),
    allow_builds=True,
    only_builds=SortedTuple(),
    prefer_older_binary=False,
    use_pep517=None,
    build_isolation=True,
    transitive=True,
    excluded=SortedTuple(),
    overridden=SortedTuple(),
    locked_resolves=SortedTuple(
        [
            LockedResolve(
                locked_requirements=SortedTuple(
                    [
                        LockedRequirement(
                            pin=Pin(ProjectName("ansicolors"), Version("1.1.8")),
                            artifact=Artifact.from_url(
                                url="http://localhost:9999/ansicolors-1.1.8-py2.py3-none-any.whl",
                                fingerprint=Fingerprint(algorithm="md5", hash="abcd1234"),
                            ),
                            additional_artifacts=SortedTuple(
                                [
                                    Artifact.from_url(
                                        url="http://localhost:9999/ansicolors-1.1.8.tar.gz",
                                        fingerprint=Fingerprint(algorithm="sha1", hash="ef567890"),
                                    )
                                ]
                            ),
                        )
                    ]
                )
            )
        ]
    ),
    local_project_requirement_mapping={},
)


def export(
    tmpdir,  # type: Any
    lockfile,  # type: Lockfile
    lockfile_path=None,  # type: Optional[str]
    export_args=(),  # type: Iterable[str]
    expected_error_re=None,  # type: Optional[str]
):
    # type: (...) -> Text
    lock = lockfile_path or os.path.join(str(tmpdir), "lock.json")
    with open(lock, "w") as fp:
        json.dump(json_codec.as_json_data(lockfile), fp, sort_keys=True, indent=2)

    result = run_pex3(*(("lock", "export", lock) + tuple(export_args)))
    result.assert_success(expected_error_re=expected_error_re)
    return result.output


def test_export_multiple_artifacts(tmpdir):
    # type: (Any) -> None

    assert (
        dedent(
            """\
            ansicolors==1.1.8 \\
              --hash=md5:abcd1234 \\
              --hash=sha1:ef567890
            """
        )
        == export(tmpdir, UNIVERSAL_ANSICOLORS)
    )


def test_export_single_artifact(tmpdir):
    # type: (Any) -> None

    assert (
        dedent(
            """\
            ansicolors==1.1.8 \\
              --hash=sha1:ef567890
            """
        )
        == export(tmpdir, attr.evolve(UNIVERSAL_ANSICOLORS, allow_wheels=False))
    )


def test_export_normalizes_name_but_not_version(tmpdir):
    # type: (Any) -> None

    assert dedent(
        """\
             twitter-common-decorators==1.3.0 \\
               --hash=md5:abcd1234 \\
               --hash=sha1:ef567890
            """
    ) == export(
        tmpdir,
        attr.evolve(
            UNIVERSAL_ANSICOLORS,
            requirements=SortedTuple([Requirement.parse("twitter.common.decorators")]),
            locked_resolves=SortedTuple(
                [
                    LockedResolve(
                        locked_requirements=SortedTuple(
                            [
                                LockedRequirement(
                                    pin=Pin(
                                        ProjectName("twitter.common.decorators"),
                                        Version("1.3.0"),
                                    ),
                                    artifact=Artifact.from_url(
                                        url="http://localhost:9999/twitter.common.decorators-1.3.0-py2.py3-none-any.whl",
                                        fingerprint=Fingerprint(algorithm="md5", hash="abcd1234"),
                                    ),
                                    additional_artifacts=SortedTuple(
                                        [
                                            Artifact.from_url(
                                                url="http://localhost:9999/twitter.common.decorators-1.3.0.tar.gz",
                                                fingerprint=Fingerprint(
                                                    algorithm="sha1", hash="ef567890"
                                                ),
                                            )
                                        ]
                                    ),
                                )
                            ],
                        ),
                    )
                ]
            ),
        ),
    )


def test_export_sort_by(tmpdir):
    # type: (Any) -> None
    ansicolors_plus_attrs = attr.evolve(
        UNIVERSAL_ANSICOLORS,
        requirements=[
            Requirement.parse("a-package"),
            Requirement.parse("z-package"),
        ],
        locked_resolves=[
            attr.evolve(
                UNIVERSAL_ANSICOLORS.locked_resolves[0],
                locked_requirements=(
                    [
                        LockedRequirement(
                            pin=Pin(ProjectName("a-package"), Version("1.1.8")),
                            artifact=Artifact.from_url(
                                url="http://localhost:9999/a-package-1.1.8-py2.py3-none-any.whl",
                                fingerprint=Fingerprint(algorithm="md5", hash="abcd1234"),
                            ),
                        ),
                        LockedRequirement(
                            pin=Pin(ProjectName("other-package"), Version("0.1.3")),
                            artifact=Artifact.from_url(
                                url="http://localhost:9999/other-package-0.1.3-py2.py3-none-any.whl",
                                fingerprint=Fingerprint(algorithm="sha256", hash="spamspam"),
                            ),
                        ),
                        LockedRequirement(
                            pin=Pin(ProjectName("z-package"), Version("22.1.0")),
                            requires_dists=SortedTuple([Requirement("other-package")]),
                            artifact=Artifact.from_url(
                                url="http://localhost:9999/z-package-22.1.0-py2.py3-none-any.whl",
                                fingerprint=Fingerprint(algorithm="sha256", hash="spameggs"),
                            ),
                        ),
                    ]
                ),
            )
        ],
    )
    assert (
        dedent(
            """\
            a-package==1.1.8 \\
              --hash=md5:abcd1234
            z-package==22.1.0 \\
              --hash=sha256:spameggs
            other-package==0.1.3 \\
              --hash=sha256:spamspam
            """
        )
        == export(tmpdir, ansicolors_plus_attrs, export_args=("--sort-by", "specificity"))
    )

    assert (
        dedent(
            """\
            a-package==1.1.8 \\
              --hash=md5:abcd1234
            other-package==0.1.3 \\
              --hash=sha256:spamspam
            z-package==22.1.0 \\
              --hash=sha256:spameggs
            """
        )
        == export(tmpdir, ansicolors_plus_attrs, export_args=("--sort-by", "project-name"))
    )


def test_export_respects_target(tmpdir):
    # type: (Any) -> None

    ansicolors_plus_pywin32 = attr.evolve(
        UNIVERSAL_ANSICOLORS,
        requirements=SortedTuple(
            [
                Requirement.parse("ansicolors"),
                Requirement.parse('pywin32; sys_platform == "win32"'),
            ]
        ),
        locked_resolves=SortedTuple(
            [
                attr.evolve(
                    UNIVERSAL_ANSICOLORS.locked_resolves[0],
                    locked_requirements=SortedTuple(
                        list(UNIVERSAL_ANSICOLORS.locked_resolves[0].locked_requirements)
                        + [
                            LockedRequirement(
                                pin=Pin(ProjectName("pywin32"), Version("227")),
                                artifact=Artifact.from_url(
                                    url="http://localhost:9999/pywin32-227-cp39-cp39-win32.whl",
                                    fingerprint=Fingerprint(algorithm="sha256", hash="spameggs"),
                                ),
                            )
                        ]
                    ),
                )
            ]
        ),
    )
    assert dedent(
        """\
            ansicolors==1.1.8 \\
              --hash=md5:abcd1234 \\
              --hash=sha1:ef567890
            pywin32==227 \\
              --hash=sha256:spameggs
            """
    ) == export(
        tmpdir,
        ansicolors_plus_pywin32,
        export_args=(
            "--complete-platform",
            json.dumps(
                {
                    "marker_environment": {"sys_platform": "win32"},
                    "compatible_tags": ["cp39-cp39-win32", "py3-none-any"],
                }
            ),
        ),
    ), (
        "A win32 foreign target should get both ansicolors cross-platform artifacts as well as "
        "the platform-specific pywin32 wheel."
    )

    assert (
        dedent(
            """\
        ansicolors==1.1.8 \\
          --hash=md5:abcd1234 \\
          --hash=sha1:ef567890
        """
        )
        == export(tmpdir, ansicolors_plus_pywin32, export_args=("--python", sys.executable))
    ), "The local interpreter doesn't support Windows; so we should just get ansicolors artifacts."


@pytest.fixture
def ansicolors_plus_vcs_plus_local_project(pex_project_dir):
    # type: (str) -> Lockfile
    return attr.evolve(
        UNIVERSAL_ANSICOLORS,
        requirements=SortedTuple(
            [
                Requirement.parse("ansicolors"),
                Requirement.parse("cowsay"),
                Requirement.parse("pex=={version}".format(version=__version__)),
            ],
            key=str,
        ),
        locked_resolves=SortedTuple(
            [
                attr.evolve(
                    UNIVERSAL_ANSICOLORS.locked_resolves[0],
                    locked_requirements=SortedTuple(
                        list(UNIVERSAL_ANSICOLORS.locked_resolves[0].locked_requirements)
                        + [
                            LockedRequirement(
                                pin=Pin(ProjectName("cowsay"), Version("6.1")),
                                artifact=VCSArtifact.from_artifact_url(
                                    artifact_url=ArtifactURL.parse(
                                        "git+https://github.com/VaasuDevanS/cowsay-python@3db622ce"
                                    ),
                                    fingerprint=Fingerprint(algorithm="sha256", hash="moo"),
                                ),
                            ),
                            LockedRequirement(
                                pin=Pin(ProjectName("pex"), Version(__version__)),
                                artifact=LocalProjectArtifact(
                                    url=ArtifactURL.parse(
                                        "file://{pex_project_dir}".format(
                                            pex_project_dir=pex_project_dir
                                        )
                                    ),
                                    fingerprint=Fingerprint(algorithm="sha256", hash="pex"),
                                    verified=False,
                                    directory=pex_project_dir,
                                ),
                            ),
                        ]
                    ),
                )
            ]
        ),
        local_project_requirement_mapping={
            pex_project_dir: Requirement.parse("pex=={version}".format(version=__version__))
        },
    )


def test_export_vcs_and_local_project_requirements_issue_2416(
    tmpdir,  # type: Any
    ansicolors_plus_vcs_plus_local_project,  # type: Lockfile
    pex_project_dir,  # type: str
):
    # type: (...) -> None

    lockfile_path = os.path.join(str(tmpdir), "lock.json")
    expected_error_msg = dedent(
        """\
        The requirements exported from {lockfile} include the following requirements
        that tools likely won't support --hash for:
        + VCS requirement 'cowsay @ git+https://github.com/VaasuDevanS/cowsay-python@3db622ce'
        + local project requirement 'pex @ file://{pex_project_dir}'

        If you can accept a lack of hash checking you can specify `--format pip-no-hashes`.
        """
    ).format(lockfile=lockfile_path, pex_project_dir=pex_project_dir)
    exported = export(
        tmpdir,
        ansicolors_plus_vcs_plus_local_project,
        lockfile_path=lockfile_path,
        expected_error_re=re.escape(expected_error_msg),
    )
    assert (
        dedent(
            """\
            ansicolors==1.1.8 \\
              --hash=md5:abcd1234 \\
              --hash=sha1:ef567890
            cowsay @ git+https://github.com/VaasuDevanS/cowsay-python@3db622ce \\
              --hash=sha256:moo
            pex @ file://{pex_project_dir} \\
              --hash=sha256:pex
            """
        ).format(pex_project_dir=pex_project_dir)
        == exported
    ), exported


def test_export_no_hashes(
    tmpdir,  # type: Any
    ansicolors_plus_vcs_plus_local_project,  # type: Lockfile
    pex_project_dir,  # type: str
):
    # type: (...) -> None

    exported = export(
        tmpdir, ansicolors_plus_vcs_plus_local_project, export_args=("--format", "pip-no-hashes")
    )
    assert (
        dedent(
            """\
            ansicolors==1.1.8
            cowsay @ git+https://github.com/VaasuDevanS/cowsay-python@3db622ce
            pex @ file://{pex_project_dir}
            """
        ).format(pex_project_dir=pex_project_dir)
        == exported
    ), exported



#!/usr/bin/env python3

from __future__ import annotations

import argparse
import atexit
import os.path
import platform
import shutil
import subprocess
import sys
import tarfile
import tempfile
from enum import Enum
from pathlib import Path, PurePath
from typing import Iterable, Optional
from urllib.parse import unquote, urlparse

import httpx

PEX_DEV_DIR = Path("~/.pex_dev").expanduser()

PAGEFIND_NAME = "pagefind"
PAGEFIND_VERSION = "1.0.4"


class Platform(Enum):
    Linux_aarch64 = "aarch64-unknown-linux-musl"
    Linux_x86_64 = "x86_64-unknown-linux-musl"
    Macos_aarch64 = "aarch64-apple-darwin"
    Macos_x86_64 = "x86_64-apple-darwin"
    Windows_x86_64 = "x86_64-pc-windows-msvc"

    @classmethod
    def parse(cls, value: str) -> Platform:
        return Platform.current() if "current" == value else Platform(value)

    @classmethod
    def current(cls) -> Platform:
        system = platform.system().lower()
        machine = platform.machine().lower()
        if system == "linux":
            if machine in ("aarch64", "arm64"):
                return cls.Linux_aarch64
            elif machine in ("amd64", "x86_64"):
                return cls.Linux_x86_64
        elif system == "darwin":
            if machine in ("aarch64", "arm64"):
                return cls.Macos_aarch64
            elif machine in ("amd64", "x86_64"):
                return cls.Macos_x86_64
        elif system == "windows" and machine in ("amd64", "x86_64"):
            return cls.Windows_x86_64

        raise ValueError(
            f"The current operating system / machine pair is not supported!: {system} / {machine}"
        )

    @property
    def extension(self):
        return ".exe" if self is self.Windows_x86_64 else ""

    def binary_name(self, binary_name: str) -> str:
        return f"{binary_name}{self.extension}"


CURRENT_PLATFORM = Platform.current()


def target_triple() -> str:
    return CURRENT_PLATFORM.value


def pagefind_executable() -> str:
    return CURRENT_PLATFORM.binary_name(
        binary_name="{name}-{version}".format(name=PAGEFIND_NAME, version=PAGEFIND_VERSION)
    )


def ensure_pagefind() -> PurePath:
    pagefind_exe = pagefind_executable()
    pagefind_exe_path = PEX_DEV_DIR / pagefind_exe
    if pagefind_exe_path.is_file() and os.access(pagefind_exe_path, os.R_OK | os.X_OK):
        return pagefind_exe_path

    tmp_dir = PEX_DEV_DIR / ".tmp"
    tmp_dir.mkdir(parents=True, exist_ok=True)
    download_dir = Path(tempfile.mktemp(prefix="download-pagefind.", dir=tmp_dir))
    atexit.register(shutil.rmtree, str(download_dir), ignore_errors=True)

    tarball_url = (
        f"https://github.com/CloudCannon/pagefind/releases/download/v{PAGEFIND_VERSION}/"
        f"{PAGEFIND_NAME}-v{PAGEFIND_VERSION}-{target_triple()}.tar.gz"
    )
    out_path = download_dir / PurePath(unquote(urlparse(tarball_url).path)).name
    out_path.parent.mkdir(parents=True, exist_ok=True)
    with httpx.stream("GET", tarball_url, follow_redirects=True) as response, out_path.open(
        "wb"
    ) as out_fp:
        for chunk in response.iter_bytes():
            out_fp.write(chunk)
    with tarfile.open(out_path) as tf:
        tf.extract(PAGEFIND_NAME, path=str(download_dir))
    (download_dir / PAGEFIND_NAME).rename(pagefind_exe_path)

    return pagefind_exe_path


def execute_pagefind(args: Iterable[str]) -> None:
    subprocess.run(args=[str(ensure_pagefind()), *args], check=True)


def execute_sphinx_build(out_base_dir: Path, builder: str, out_dir: Optional[str] = None) -> Path:
    gen_dir = (out_base_dir / (out_dir or builder)).absolute()
    shutil.rmtree(gen_dir, ignore_errors=True)
    subprocess.run(
        args=[sys.executable, "-m", "sphinx", "-b", builder, "-aEW", "docs", str(gen_dir)],
        check=True,
    )
    return gen_dir


def main(
    out_dir: Path,
    linkcheck: bool = False,
    pdf: bool = False,
    html: bool = True,
    clean_html: bool = False,
    serve: bool = False,
) -> None:
    static_dynamic_dir = (Path("docs") / "_static_dynamic").absolute()

    def clean_static_dynmaic_dir() -> None:
        shutil.rmtree(static_dynamic_dir, ignore_errors=True)

    clean_static_dynmaic_dir()
    static_dynamic_dir.mkdir(parents=True, exist_ok=True)
    atexit.register(clean_static_dynmaic_dir)

    if linkcheck:
        execute_sphinx_build(out_dir, "linkcheck")

    if pdf:
        pdf_dir = execute_sphinx_build(out_dir, "simplepdf", out_dir="pdf")
        (static_dynamic_dir / "pex.pdf").symlink_to(pdf_dir / "pex.pdf")

    if html:
        html_dir = execute_sphinx_build(out_dir, "html")
        if clean_html:
            shutil.rmtree(html_dir / ".doctrees", ignore_errors=True)
            (html_dir / ".buildinfo").unlink(missing_ok=True)
            (html_dir / "objects.inv").unlink(missing_ok=True)

        page_find_args = ["--site", str(html_dir), "--output-subdir", "_pagefind"]
        if serve:
            page_find_args.append("--serve")
        execute_pagefind(page_find_args)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--linkcheck", action="store_true")
    parser.add_argument("--pdf", action="store_true")
    parser.add_argument("--no-html", action="store_true")
    parser.add_argument("--clean-html", action="store_true")
    parser.add_argument("--serve", action="store_true")
    parser.add_argument("out_dir", type=Path, default=Path("dist") / "docs", nargs="?")
    options = parser.parse_args()
    try:
        main(
            out_dir=options.out_dir,
            linkcheck=options.linkcheck,
            pdf=options.pdf,
            html=not options.no_html,
            clean_html=options.clean_html,
            serve=options.serve,
        )
    except subprocess.CalledProcessError as e:
        sys.exit(e.returncode)



from __future__ import absolute_import

import json
import logging
import os
import subprocess
from contextlib import contextmanager
from textwrap import dedent

from pex.atomic_directory import atomic_directory
from pex.common import safe_rmtree
from pex.interpreter import PythonInterpreter
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import InvalidVirtualenvError, Virtualenv
from testing import PEX_TEST_DEV_ROOT, PY310, data, ensure_python_interpreter

if TYPE_CHECKING:
    from typing import Iterable, Iterator, Optional, Tuple

    import attr  # vendor:skip
else:
    from pex.third_party import attr


logger = logging.getLogger(__name__)


MITMPROXY_DIR = os.path.join(PEX_TEST_DEV_ROOT, "mitmproxy")


def _ensure_mitmproxy_venv():
    # type: () -> Virtualenv

    venv_dir = os.path.join(MITMPROXY_DIR, "venv")
    try:
        return Virtualenv(venv_dir=venv_dir)
    except InvalidVirtualenvError as e:
        logger.warning(str(e))
        safe_rmtree(venv_dir)
        with atomic_directory(venv_dir) as atomic_venvdir:
            if not atomic_venvdir.is_finalized():
                logger.info("Installing mitmproxy...")
                mitmproxy_lock = data.path("locks", "mitmproxy.lock.json")
                python = ensure_python_interpreter(PY310)
                Virtualenv.create_atomic(
                    venv_dir=atomic_venvdir,
                    interpreter=PythonInterpreter.from_binary(python),
                    force=True,
                )
                subprocess.check_call(
                    args=[
                        python,
                        "-m",
                        "pex.cli",
                        "venv",
                        "create",
                        "--lock",
                        mitmproxy_lock,
                        "-d",
                        atomic_venvdir.work_dir,
                    ]
                )
        return Virtualenv(venv_dir=venv_dir)


@attr.s(frozen=True)
class Proxy(object):
    @classmethod
    def configured(cls, config_dir):
        # type: (str) -> Proxy

        mitmdump_venv = _ensure_mitmproxy_venv()

        confdir = os.path.join(config_dir, "confdir")
        messages = os.path.join(config_dir, "messages")
        addon = os.path.join(config_dir, "addon.py")
        with open(addon, "w") as fp:
            fp.write(
                dedent(
                    """\
                    import json

                    from mitmproxy import ctx


                    def running() -> None:
                        port = ctx.master.addons.get("proxyserver").listen_addrs()[0][1]
                        with open({msg_channel!r}, "w") as fp:
                            json.dump({{"port": port}}, fp)
                    """.format(
                        msg_channel=messages
                    )
                )
            )
        return cls(mitmdump_venv=mitmdump_venv, confdir=confdir, messages=messages, addon=addon)

    mitmdump_venv = attr.ib()  # type Virtualenv
    confdir = attr.ib()  # type: str
    messages = attr.ib()  # type: str
    addon = attr.ib()  # type: str

    @contextmanager
    def reverse(
        self,
        targets,  # type: Iterable[str]
        proxy_auth=None,  # type: Optional[str]
    ):
        # type: (...) -> Iterator[Tuple[int, str]]
        os.mkfifo(self.messages)
        args = [
            self.mitmdump_venv.interpreter.binary,
            self.mitmdump_venv.bin_path("mitmdump"),
            "--set",
            "confdir={confdir}".format(confdir=self.confdir),
            "-p",
            "0",
            "-s",
            self.addon,
        ]
        if proxy_auth:
            args.extend(["--proxyauth", proxy_auth])
        for target in targets:
            args.extend(["--mode", "reverse:{target}".format(target=target)])
        proxy_process = subprocess.Popen(args)
        try:
            with open(self.messages, "r") as fp:
                data = json.load(fp)
            yield data["port"], os.path.join(self.confdir, "mitmproxy-ca.pem")
        finally:
            proxy_process.kill()
            os.unlink(self.messages)

    @contextmanager
    def run(self, proxy_auth=None):
        # type: (Optional[str]) -> Iterator[Tuple[int, str]]

        with self.reverse(targets=(), proxy_auth=proxy_auth) as (port, cert):
            yield port, cert



# Copyright 2022 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path

import pytest

from pex.layout import Layout
from pex.pep_427 import InstallableType
from pex.pex_info import PexInfo
from pex.typing import TYPE_CHECKING
from testing import run_pex_command
from testing.pep_427 import get_installable_type_flag

if TYPE_CHECKING:
    from typing import Any


# N.B.: The ordering of decorators is just so to get the test ids to match with what the test does
# for sanity's sake.
#
# So, we get "test_overwrite[zipapp-loose]" which indicates a test of the transition from
# zipapp (layout1) to loose (layout2) and "test_overwrite[packed-packed]" to indicate an overwrite
# of the packed layout by another packed layout, etc.
@pytest.mark.parametrize(
    "installable_type2",
    [
        pytest.param(installable_type, id=installable_type.value)
        for installable_type in InstallableType.values()
    ],
)
@pytest.mark.parametrize(
    "layout2", [pytest.param(layout, id=layout.value) for layout in Layout.values()]
)
@pytest.mark.parametrize(
    "installable_type1",
    [
        pytest.param(installable_type, id=installable_type.value)
        for installable_type in InstallableType.values()
    ],
)
@pytest.mark.parametrize(
    "layout1", [pytest.param(layout, id=layout.value) for layout in Layout.values()]
)
def test_overwrite(
    tmpdir,  # type: Any
    layout1,  # type: Layout.Value
    installable_type1,  # type: InstallableType.Value
    layout2,  # type: Layout.Value
    installable_type2,  # type: InstallableType.Value
):
    # type: (...) -> None

    pex = os.path.join(str(tmpdir), "pex")

    run_pex_command(
        args=[
            "-e",
            "one",
            "-o",
            pex,
            "--layout",
            layout1.value,
            get_installable_type_flag(installable_type1),
        ]
    ).assert_success()
    assert layout1 is Layout.identify(pex)
    assert "one" == PexInfo.from_pex(pex).entry_point

    run_pex_command(
        args=[
            "-e",
            "two",
            "-o",
            pex,
            "--layout",
            layout2.value,
            get_installable_type_flag(installable_type2),
        ]
    ).assert_success()
    assert layout2 is Layout.identify(pex)
    assert "two" == PexInfo.from_pex(pex).entry_point



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).
import json
import os
import subprocess
from textwrap import dedent

from pex.interpreter import PythonInterpreter
from pex.pep_425 import CompatibilityTags
from pex.pep_508 import MarkerEnvironment
from pex.pip.installation import compatible_version
from pex.pip.version import PipVersion
from pex.result import try_
from pex.targets import CompletePlatform, Targets
from pex.typing import TYPE_CHECKING
from testing import IS_X86_64, run_pex_command
from testing.docker import skip_unless_docker

if TYPE_CHECKING:
    from typing import Any


@skip_unless_docker
def test_ssl_context(
    tmpdir,  # type: Any
    pex_project_dir,  # type: str
):
    # type: (...) -> None

    with open(os.path.join(str(tmpdir), "Dockerfile"), "w") as fp:
        fp.write(
            dedent(
                r"""
                FROM fedora:37
                
                ARG PBS_RELEASE
                ARG PBS_ARCHIVE

                RUN \
                curl --fail -sSL -O $PBS_RELEASE/$PBS_ARCHIVE && \
                curl --fail -sSL -O $PBS_RELEASE/$PBS_ARCHIVE.sha256 && \
                [[ \
                    "$(cat $PBS_ARCHIVE.sha256)" == "$(sha256sum $PBS_ARCHIVE | cut -d' ' -f1)" \
                ]] && \
                tar -xzf $PBS_ARCHIVE
                """
            )
        )

    pbs_release = "https://github.com/indygreg/python-build-standalone/releases/download/20240107"
    pbs_archive = "cpython-3.9.18+20240107-{arch}-unknown-linux-gnu-install_only.tar.gz".format(
        arch="x86_64" if IS_X86_64 else "aarch64"
    )
    subprocess.check_call(
        args=[
            "docker",
            "build",
            "-t",
            "test_issue_2355",
            "--build-arg",
            "PBS_RELEASE={pbs_release}".format(pbs_release=pbs_release),
            "--build-arg",
            "PBS_ARCHIVE={pbs_archive}".format(pbs_archive=pbs_archive),
            str(tmpdir),
        ]
    )

    work_dir = os.path.join(str(tmpdir), "work_dir")
    os.mkdir(work_dir)
    docker_run_args = [
        "docker",
        "run",
        "--rm",
        "-v",
        "{pex_project_dir}:/code".format(pex_project_dir=pex_project_dir),
        "-v",
        "{work_dir}:/work".format(work_dir=work_dir),
        "-w",
        "/code",
        "test_issue_2355",
        "/python/bin/python3.9",
        "-mpex.cli",
    ]
    complete_platform_data = json.loads(
        subprocess.check_output(
            args=docker_run_args + ["interpreter", "inspect", "--markers", "--tags"]
        )
    )
    container_interpreter = CompletePlatform.create(
        marker_environment=MarkerEnvironment(**complete_platform_data["marker_environment"]),
        supported_tags=CompatibilityTags.from_strings(complete_platform_data["compatible_tags"]),
    )
    pip_version = try_(
        compatible_version(
            targets=Targets(
                interpreters=tuple([PythonInterpreter.get()]),
                complete_platforms=tuple([container_interpreter]),
            ),
            requested_version=PipVersion.DEFAULT,
            context=__name__ + ".test_ssl_context",
        )
    )

    subprocess.check_call(
        args=docker_run_args
        + [
            "lock",
            "create",
            "--pip-version",
            str(pip_version),
            "--style",
            "universal",
            "cowsay==5.0",
            "--indent",
            "2",
            "-o",
            "/work/lock.json",
        ]
    )

    result = run_pex_command(
        args=["--lock", os.path.join(work_dir, "lock.json"), "-c", "cowsay", "--", "Moo!"]
    )
    result.assert_success()
    assert "Moo!" in result.error



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import os.path
import subprocess
import sys

import pex_build
from hatchling.builders.hooks.plugin.interface import BuildHookInterface

if pex_build.TYPE_CHECKING:
    from typing import Any, Dict


class AdjustBuild(BuildHookInterface):
    """Allows alteration of the build process."""

    PLUGIN_NAME = "pex-adjust-build"

    def initialize(
        self,
        version,  # type: str
        build_data,  # type: Dict[str, Any]
    ):
        # type: (...) -> None
        if pex_build.INCLUDE_DOCS:
            out_dir = os.path.join(self.root, "dist", "docs")
            subprocess.check_call(
                args=[
                    sys.executable,
                    os.path.join(self.root, "scripts", "build_docs.py"),
                    "--clean-html",
                    out_dir,
                ]
            )
            build_data["force_include"][out_dir] = os.path.join("pex", "docs")



# Copyright 2020 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import logging
import os
import pkgutil
import re
import shutil
import sys
from collections import defaultdict
from contextlib import closing
from fileinput import FileInput
from textwrap import dedent

from pex.atomic_directory import AtomicDirectory, atomic_directory
from pex.common import is_exe, safe_mkdir, safe_open
from pex.compatibility import commonpath, get_stdout_bytes_buffer
from pex.dist_metadata import Distribution, find_distributions
from pex.enum import Enum
from pex.executor import Executor
from pex.fetcher import URLFetcher
from pex.interpreter import (
    Platlib,
    Purelib,
    PythonInterpreter,
    PyVenvCfg,
    SitePackagesDir,
    create_shebang,
)
from pex.orderedset import OrderedSet
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING, cast
from pex.util import named_temporary_file
from pex.variables import ENV
from pex.version import __version__

if TYPE_CHECKING:
    from typing import DefaultDict, Iterable, Iterator, Optional, Tuple, Type, Union

logger = logging.getLogger(__name__)


class PipUnavailableError(Exception):
    """Indicates no local copy of Pip could be found for install."""


def _iter_files(directory):
    # type: (str) -> Iterator[str]
    for entry in os.listdir(directory):
        yield os.path.join(directory, entry)


def _is_python_script(executable):
    # type: (str) -> bool
    with open(executable, "rb") as fp:
        if fp.read(2) != b"#!":
            return False
        interpreter = fp.readline()
        return bool(
            # Support the `#!python` shebang that wheel installers should recognize as a special
            # form to convert to a localized shebang upon install.
            # See: https://www.python.org/dev/peps/pep-0427/#recommended-installer-features
            interpreter == b"python\n"
            or re.search(
                br"""
                # The aim is to admit the common shebang forms:
                # + /usr/bin/env <python bin name>
                # + /absolute/path/to/<python bin name>
                \W

                # Python executable names Pex supports (see PythonIdentity).
                (
                      python
                    | pypy
                )
                # Optional Python version
                (\d+(\.\d+)*)?

                ([^.a-zA-Z0-9]|$)
                """,
                interpreter,
                re.VERBOSE,
            )
        )


class InvalidVirtualenvError(Exception):
    """Indicates a virtualenv is malformed."""


def _find_preferred_site_packages_dir(
    venv_dir,  # type: str
    interpreter=None,  # type: Optional[PythonInterpreter]
):
    # type: (...) -> str

    real_venv_dir = os.path.realpath(venv_dir)
    site_packages_dirs_by_type = defaultdict(
        OrderedSet
    )  # type: DefaultDict[Type[SitePackagesDir], OrderedSet[SitePackagesDir]]

    interpreter = interpreter or PythonInterpreter.get()
    for entry in interpreter.site_packages:
        if commonpath((real_venv_dir, entry.path)) != real_venv_dir:
            # This ignores system site packages when the venv is built with --system-site-packages.
            continue
        if os.path.isdir(entry.path):
            site_packages_dirs_by_type[type(entry)].add(entry)

    if not site_packages_dirs_by_type:
        raise InvalidVirtualenvError(
            "The virtualenv at {venv_dir} is not valid. No site-packages directory was found in "
            "its sys.path:\n{sys_path}".format(
                venv_dir=venv_dir, sys_path="\n".join(interpreter.sys_path)
            )
        )

    for site_packages_dir_type in Purelib, Platlib, SitePackagesDir:
        site_packages_dirs = site_packages_dirs_by_type.get(site_packages_dir_type)
        if not site_packages_dirs:
            continue
        if len(site_packages_dirs) > 1:
            raise InvalidVirtualenvError(
                "The virtualenv at {venv_dir} is not valid. It has more than one {dir_type} "
                "directory:\n{site_packages}".format(
                    venv_dir=venv_dir,
                    dir_type=site_packages_dir_type,
                    site_packages="\n".join(entry.path for entry in site_packages_dirs),
                )
            )
        return site_packages_dirs.pop().path

    raise InvalidVirtualenvError(
        "Could not determine the site-packages directory for the venv at {venv_dir}.".format(
            venv_dir=venv_dir
        )
    )


class InstallationChoice(Enum["InstallationChoice.Value"]):
    class Value(Enum.Value):
        pass

    NO = Value("no")
    YES = Value("yes")
    UPGRADED = Value("upgraded")


class Virtualenv(object):
    VIRTUALENV_VERSION = "16.7.12"

    @classmethod
    def enclosing(cls, python):
        # type: (Union[str, PythonInterpreter]) -> Optional[Virtualenv]
        """Return the virtual environment the given python interpreter is enclosed in."""
        if isinstance(python, PythonInterpreter):
            interpreter = python
        else:
            try:
                interpreter = PythonInterpreter.from_binary(python)
            except PythonInterpreter.Error:
                return None

        if not interpreter.is_venv:
            return None
        return cls(
            venv_dir=interpreter.prefix, python_exe_name=os.path.basename(interpreter.binary)
        )

    @classmethod
    def create(
        cls,
        venv_dir,  # type: str
        interpreter=None,  # type: Optional[PythonInterpreter]
        force=False,  # type: bool
        copies=False,  # type: bool
        system_site_packages=False,  # type: bool
        prompt=None,  # type: Optional[str]
        install_pip=InstallationChoice.NO,  # type: InstallationChoice.Value
        install_setuptools=InstallationChoice.NO,  # type: InstallationChoice.Value
        install_wheel=InstallationChoice.NO,  # type: InstallationChoice.Value
        other_installs=(),  # type: Iterable[str]
    ):
        # type: (...) -> Virtualenv

        installations = {
            "pip": install_pip,
            "setuptools": install_setuptools,
            "wheel": install_wheel,
        }
        project_upgrades = [
            project
            for project, installation_choice in installations.items()
            if installation_choice is InstallationChoice.UPGRADED
        ]
        if project_upgrades and install_pip is InstallationChoice.NO:
            raise ValueError(
                "Installation of Pip is required in order to upgrade {projects}.".format(
                    projects=" and ".join(project_upgrades)
                )
            )

        venv_dir = os.path.abspath(venv_dir)
        safe_mkdir(venv_dir, clean=force)

        interpreter = interpreter or PythonInterpreter.get()
        if interpreter.is_venv:
            base_interpreter = interpreter.resolve_base_interpreter()
            TRACER.log(
                "Ignoring enclosing venv {} and using its base interpreter {} to create venv at {}"
                " instead.".format(interpreter.prefix, base_interpreter.binary, venv_dir),
                V=3,
            )
            interpreter = base_interpreter

        # N.B.: PyPy3.6 and PyPy3.7 come equipped with a venv module but it does not seem to
        # work.
        py_major_minor = interpreter.version[:2]
        use_virtualenv = py_major_minor[0] == 2 or (
            interpreter.is_pypy and py_major_minor[:2] <= (3, 7)
        )

        installations.pop("pip")
        if install_pip is not InstallationChoice.NO and py_major_minor < (3, 12):
            # The ensure_pip module, get_pip.py and the venv module all install setuptools when
            # they install Pip for all Pythons older than 3.12.
            installations.pop("setuptools")

        project_installs = OrderedSet(
            project
            for project, installation_choice in installations.items()
            if installation_choice is InstallationChoice.YES
        )
        project_installs.update(other_installs)
        if project_installs and install_pip is InstallationChoice.NO:
            raise ValueError(
                "Installation of Pip is required in order to install {projects}.".format(
                    projects=" and ".join(project_installs)
                )
            )

        # Guard against API calls from environment with ambient PYTHONPATH preventing pip virtualenv
        # creation. See: https://github.com/pex-tool/pex/issues/1451
        env = os.environ.copy()
        pythonpath = env.pop("PYTHONPATH", None)
        if pythonpath:
            TRACER.log(
                "Scrubbed PYTHONPATH={} from the virtualenv creation environment.".format(
                    pythonpath
                ),
                V=3,
            )

        custom_prompt = None  # type: Optional[str]
        if use_virtualenv:
            virtualenv_py = pkgutil.get_data(
                __name__, "virtualenv_{version}_py".format(version=cls.VIRTUALENV_VERSION)
            )
            with named_temporary_file(mode="wb") as fp:
                fp.write(virtualenv_py)
                fp.close()
                args = [fp.name, "--no-pip", "--no-setuptools", "--no-wheel", venv_dir]
                if copies:
                    args.append("--always-copy")
                if system_site_packages:
                    args.append("--system-site-packages")
                if prompt:
                    args.extend(["--prompt", prompt])
                    custom_prompt = prompt
                interpreter.execute(args=args, env=env)
                # Modern virtualenv provides a pyvenv.cfg; so we provide one on 16.7.12's behalf
                # since users might expect one. To ward off any confusion for readers of the emitted
                # pyvenv.cfg file, we add a bespoke created-by field to help make it clear that Pex
                # created the pyvenv.cfg file on virtualenv 16.7.12's behalf.
                # N.B.: This bespoke created-by "note" field is not related to the
                # Virtualenv.created_by property which reflects the underlying venv technology.
                # In this case it will report "virtualenv 16.7.12".
                with open(os.path.join(venv_dir, "pyvenv.cfg"), "w") as fp:
                    fp.write(
                        dedent(
                            """\
                            home = {home}
                            include-system-site-packages = {include_system_site_packages}
                            virtualenv = {virtualenv_version}
                            version = {python_version}
                            created-by = pex {pex_version}
                            """
                        ).format(
                            home=os.path.dirname(interpreter.binary),
                            include_system_site_packages=(
                                "true" if system_site_packages else "false"
                            ),
                            virtualenv_version=cls.VIRTUALENV_VERSION,
                            python_version=".".join(map(str, interpreter.version)),
                            pex_version=__version__,
                        )
                    )
        else:
            args = ["-m", "venv", venv_dir]
            if install_pip is InstallationChoice.NO:
                args.append("--without-pip")
            if copies:
                args.append("--copies")
            if system_site_packages:
                args.append("--system-site-packages")
            if prompt and py_major_minor >= (3, 6):
                args.extend(["--prompt", prompt])
                custom_prompt = prompt
            interpreter.execute(args=args, env=env)

        venv = cls(venv_dir, custom_prompt=custom_prompt)
        if use_virtualenv and (
            install_pip is not InstallationChoice.NO or project_upgrades or project_installs
        ):
            # Our vendored virtualenv does not support installing Pip, setuptool or wheel; so we
            # use the ensurepip module / get_pip.py bootstrapping for Pip that `ensure_pip` does.
            venv.ensure_pip(upgrade=install_pip is InstallationChoice.UPGRADED)
        if project_upgrades:
            venv.interpreter.execute(
                args=["-m", "pip", "install", "-U"] + project_upgrades, env=env
            )
        if project_installs:
            venv.interpreter.execute(
                args=["-m", "pip", "install"] + list(project_installs), env=env
            )
        return venv

    @classmethod
    def create_atomic(
        cls,
        venv_dir,  # type: AtomicDirectory
        interpreter=None,  # type: Optional[PythonInterpreter]
        force=False,  # type: bool
        copies=False,  # type: bool
        prompt=None,  # type: Optional[str]
        install_pip=InstallationChoice.NO,  # type: InstallationChoice.Value
        install_setuptools=InstallationChoice.NO,  # type: InstallationChoice.Value
        install_wheel=InstallationChoice.NO,  # type: InstallationChoice.Value
        other_installs=(),  # type: Iterable[str]
    ):
        # type: (...) -> Virtualenv
        virtualenv = cls.create(
            venv_dir=venv_dir.work_dir,
            interpreter=interpreter,
            force=force,
            copies=copies,
            prompt=prompt,
            install_pip=install_pip,
            install_setuptools=install_setuptools,
            install_wheel=install_wheel,
            other_installs=other_installs,
        )
        for script in virtualenv._rewrite_base_scripts(real_venv_dir=venv_dir.target_dir):
            TRACER.log("Re-writing {}".format(script))

        # It's known that PyPy's 7.3.14 releases create venvs with absolute symlinks in bin/ to
        # bin/ local files, which leaves invalid symlinks to the atomic workdir. We fix that up
        # here. See: https://github.com/pypy/pypy/issues/4838
        for path in os.listdir(virtualenv.bin_dir):
            abs_path = os.path.join(virtualenv.bin_dir, path)
            if not os.path.islink(abs_path):
                continue
            link_target = os.readlink(abs_path)
            if not os.path.isabs(link_target):
                continue
            if virtualenv.bin_dir == commonpath((virtualenv.bin_dir, link_target)):
                rel_dst = os.path.relpath(link_target, virtualenv.bin_dir)
                TRACER.log(
                    "Replacing absolute symlink {src} -> {dst} with relative symlink".format(
                        src=abs_path, dst=link_target
                    ),
                    V=3,
                )
                os.unlink(abs_path)
                os.symlink(rel_dst, abs_path)

        return virtualenv

    def __init__(
        self,
        venv_dir,  # type: str
        python_exe_name="python",  # type: str
        custom_prompt=None,  # type: Optional[str]
    ):
        # type: (...) -> None
        self._venv_dir = venv_dir
        self._custom_prompt = custom_prompt
        self._bin_dir = os.path.join(venv_dir, "bin")
        python_exe_path = os.path.join(self._bin_dir, python_exe_name)
        try:
            self._interpreter = PythonInterpreter.from_binary(python_exe_path)
        except PythonInterpreter.Error as e:
            raise InvalidVirtualenvError(
                "The virtualenv at {venv_dir} is not valid. Failed to load an interpreter at "
                "{python_exe_path}: {err}".format(
                    venv_dir=self._venv_dir, python_exe_path=python_exe_path, err=e
                )
            )
        self._site_packages_dir = _find_preferred_site_packages_dir(venv_dir, self._interpreter)
        self._purelib = self._site_packages_dir
        self._platlib = self._site_packages_dir
        for entry in self._interpreter.site_packages:
            if isinstance(entry, Purelib):
                self._purelib = entry.path
            elif isinstance(entry, Platlib):
                self._platlib = entry.path

        self._sys_path = None  # type: Optional[Tuple[str, ...]]

    @property
    def venv_dir(self):
        # type: () -> str
        return self._venv_dir

    @property
    def _pyvenv_cfg(self):
        # type: () -> Optional[PyVenvCfg]
        if not hasattr(self, "__pyvenv_cfg"):
            self.__pyvenv_cfg = PyVenvCfg.find(self._interpreter.binary)
        return self.__pyvenv_cfg

    @property
    def created_by(self):
        # type: () -> str
        if self._pyvenv_cfg:
            version = self._pyvenv_cfg.config("virtualenv", None)
            return "virtualenv {version}".format(version=version) if version else "venv"
        return "unknown"

    @property
    def include_system_site_packages(self):
        # type: () -> Optional[bool]
        return self._pyvenv_cfg.include_system_site_packages if self._pyvenv_cfg else None

    @property
    def custom_prompt(self):
        # type: () -> Optional[str]
        return self._custom_prompt

    def join_path(self, *components):
        # type: (*str) -> str
        return os.path.join(self._venv_dir, *components)

    def bin_path(self, *components):
        # type: (*str) -> str
        return os.path.join(self._bin_dir, *components)

    @property
    def bin_dir(self):
        # type: () -> str
        return self._bin_dir

    @property
    def site_packages_dir(self):
        # type: () -> str
        return self._site_packages_dir

    @property
    def purelib(self):
        # type: () -> str
        return self._purelib

    @property
    def platlib(self):
        # type: () -> str
        return self._platlib

    @property
    def interpreter(self):
        # type: () -> PythonInterpreter
        return self._interpreter

    def iter_executables(self):
        # type: () -> Iterator[str]
        for path in _iter_files(self._bin_dir):
            if is_exe(path):
                yield path

    @property
    def sys_path(self):
        # type: () -> Tuple[str, ...]
        return self.interpreter.sys_path

    def iter_distributions(self, rescan=False):
        # type: (bool) -> Iterator[Distribution]
        for dist in find_distributions(
            search_path=[entry.path for entry in self._interpreter.site_packages], rescan=rescan
        ):
            yield dist

    def _rewrite_base_scripts(self, real_venv_dir):
        # type: (str) -> Iterator[str]
        scripts = [
            path
            for path in _iter_files(self._bin_dir)
            if _is_python_script(path) or re.search(r"^[Aa]ctivate", os.path.basename(path))
        ]
        if scripts:
            rewritten_files = set()
            with closing(FileInput(files=sorted(scripts), inplace=True)) as fi:
                for line in fi:
                    rewritten_line = line.replace(self._venv_dir, real_venv_dir)
                    if rewritten_line != line:
                        filename = fi.filename()
                        if filename not in rewritten_files:
                            rewritten_files.add(filename)
                            yield filename
                    sys.stdout.write(rewritten_line)

    def rewrite_scripts(
        self,
        python=None,  # type: Optional[str]
        python_args=None,  # type: Optional[str]
    ):
        # type: (...) -> Iterator[str]
        python_scripts = [
            executable for executable in self.iter_executables() if _is_python_script(executable)
        ]
        if python_scripts:
            with closing(FileInput(files=sorted(python_scripts), inplace=True, mode="rb")) as fi:
                # N.B.: `FileInput` is strange, but useful: the context manager above monkey-patches
                # sys.stdout to print to the corresponding original input file, which is has moved
                # aside.
                for line in fi:
                    buffer = get_stdout_bytes_buffer()
                    if fi.isfirstline():
                        shebang = create_shebang(
                            python_exe=python or self._interpreter.binary, python_args=python_args
                        )
                        buffer.write("{shebang}\n".format(shebang=shebang).encode("utf-8"))
                        yield fi.filename()
                    else:
                        # N.B.: These lines include the newline already.
                        buffer.write(cast(bytes, line))

    def ensure_pip(self, upgrade=False):
        # type: (bool) -> str
        pip_script = self.bin_path("pip")
        if is_exe(pip_script) and not upgrade:
            return pip_script
        try:
            self._interpreter.execute(args=["-m", "ensurepip", "-U", "--default-pip"])
        except Executor.NonZeroExit:
            # Early Python 2.7 versions and some system Pythons do not come with ensurepip
            # installed. We fall back to get-pip.py which is available in dedicated versions for
            # Python 2.{6,7} and 3.{2,3,4,5,6} and a single version for anything newer.
            get_pip_script = "get-pip.py"
            major, minor = self._interpreter.version[:2]
            if (major, minor) <= (3, 6):
                version_dir = "{major}.{minor}".format(major=major, minor=minor)
                url_rel_path = "{version_dir}/{script}".format(
                    version_dir=version_dir, script=get_pip_script
                )
                dst_rel_path = os.path.join(version_dir, get_pip_script)
            else:
                url_rel_path = get_pip_script
                dst_rel_path = os.path.join("default", get_pip_script)
            get_pip = os.path.join(ENV.PEX_ROOT, "get-pip", dst_rel_path)
            with atomic_directory(os.path.dirname(get_pip)) as atomic_dir:
                if not atomic_dir.is_finalized():
                    with URLFetcher().get_body_stream(
                        "https://bootstrap.pypa.io/pip/" + url_rel_path
                    ) as src_fp, safe_open(
                        os.path.join(atomic_dir.work_dir, os.path.basename(get_pip)), "wb"
                    ) as dst_fp:
                        shutil.copyfileobj(src_fp, dst_fp)
            self._interpreter.execute(args=[get_pip, "--no-wheel"])
        if upgrade:
            self._interpreter.execute(args=["-m", "pip", "install", "-U", "pip"])
        return pip_script



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import subprocess
import sys

import pytest

from pex import layout
from pex.common import open_zip, touch
from pex.typing import TYPE_CHECKING
from testing import run_pex_command

if TYPE_CHECKING:
    from typing import Any, Iterable, Optional

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class SourceTree(object):
    base_dir = attr.ib()  # type: str

    def create(self, offset=None):
        # type: (Optional[str]) -> str

        project_dir = os.path.join(self.base_dir, "project")
        source_root = os.path.join(project_dir, offset) if offset else project_dir

        touch(os.path.join(source_root, "top_level_module.py"))

        touch(os.path.join(source_root, "top_level_package_classic", "__init__.py"))
        touch(os.path.join(source_root, "top_level_package_classic", "module.py"))
        touch(os.path.join(source_root, "top_level_package_classic", "sub", "__init__.py"))
        touch(os.path.join(source_root, "top_level_package_classic", "sub", "foo.py"))
        touch(os.path.join(source_root, "top_level_package_classic", "sub", "bar.py"))

        # N.B.: This test will run against a range of interpreters, some supporting PEP420 and some
        # not, so we never import code in the tests that use this fixture and instead just check
        # file lists,
        touch(os.path.join(source_root, "top_level_package_pep420", "module.py"))
        touch(os.path.join(source_root, "top_level_package_pep420", "sub", "module.py"))

        touch(os.path.join(source_root, "top_level_package_mixed", "module.py"))
        touch(os.path.join(source_root, "top_level_package_mixed", "sub", "__init__.py"))
        touch(os.path.join(source_root, "top_level_package_mixed", "sub", "module.py"))

        subprocess.check_call(args=[sys.executable, "-m", "compileall", source_root])

        return project_dir

    def assert_sources(
        self,
        pex_args,  # type: Iterable[str]
        expected_sources,  # type: Iterable[str]
        offset=None,  # type: Optional[str]
    ):
        # type: (...) -> None
        pex = os.path.join(self.base_dir, "pex")
        args = ["-o", pex]
        args.extend(pex_args)
        run_pex_command(args=args, cwd=self.create(offset=offset)).assert_success()

        with open_zip(pex) as zf:
            actual_sources = [
                f
                for f in zf.namelist()
                if not f.startswith((layout.BOOTSTRAP_DIR, layout.DEPS_DIR, "__pex__"))
                and f not in (layout.PEX_INFO_PATH, "__main__.py")
            ]

        assert sorted(expected_sources) == sorted(actual_sources)


@pytest.fixture
def source_tree(tmpdir):
    # type: (Any) -> SourceTree
    return SourceTree(str(tmpdir))


def test_add_top_level_package(source_tree):
    # type: (SourceTree) -> None
    source_tree.assert_sources(
        pex_args=["-P", "top_level_package_classic"],
        expected_sources=[
            "top_level_package_classic/",
            "top_level_package_classic/__init__.py",
            "top_level_package_classic/module.py",
            "top_level_package_classic/sub/",
            "top_level_package_classic/sub/__init__.py",
            "top_level_package_classic/sub/foo.py",
            "top_level_package_classic/sub/bar.py",
        ],
    )


def test_add_sub_package(source_tree):
    # type: (SourceTree) -> None
    source_tree.assert_sources(
        pex_args=["-P", "top_level_package_classic.sub"],
        expected_sources=[
            "top_level_package_classic/",
            "top_level_package_classic/__init__.py",
            "top_level_package_classic/sub/",
            "top_level_package_classic/sub/__init__.py",
            "top_level_package_classic/sub/foo.py",
            "top_level_package_classic/sub/bar.py",
        ],
    )


def test_add_sub_package_pep_420(source_tree):
    # type: (SourceTree) -> None
    source_tree.assert_sources(
        pex_args=["-P", "top_level_package_pep420.sub"],
        expected_sources=[
            "top_level_package_pep420/",
            "top_level_package_pep420/sub/",
            "top_level_package_pep420/sub/module.py",
        ],
    )


def test_add_sub_package_mixed(source_tree):
    # type: (SourceTree) -> None
    source_tree.assert_sources(
        pex_args=["-P", "top_level_package_mixed.sub"],
        expected_sources=[
            "top_level_package_mixed/",
            "top_level_package_mixed/sub/",
            "top_level_package_mixed/sub/__init__.py",
            "top_level_package_mixed/sub/module.py",
        ],
    )


def test_add_package_offset(source_tree):
    # type: (SourceTree) -> None
    source_tree.assert_sources(
        pex_args=["-P", "top_level_package_classic.sub@src"],
        offset="src",
        expected_sources=[
            "top_level_package_classic/",
            "top_level_package_classic/__init__.py",
            "top_level_package_classic/sub/",
            "top_level_package_classic/sub/__init__.py",
            "top_level_package_classic/sub/foo.py",
            "top_level_package_classic/sub/bar.py",
        ],
    )


def test_add_top_level_module(source_tree):
    # type: (SourceTree) -> None
    source_tree.assert_sources(
        pex_args=["-M", "top_level_module"], expected_sources=["top_level_module.py"]
    )


def test_add_module_in_package(source_tree):
    # type: (SourceTree) -> None
    source_tree.assert_sources(
        pex_args=["-M", "top_level_package_classic.sub.foo"],
        expected_sources=[
            "top_level_package_classic/",
            "top_level_package_classic/__init__.py",
            "top_level_package_classic/sub/",
            "top_level_package_classic/sub/__init__.py",
            "top_level_package_classic/sub/foo.py",
        ],
    )


def test_add_module_offset(source_tree):
    # type: (SourceTree) -> None
    offset = os.path.join("src", "python")
    source_tree.assert_sources(
        pex_args=["-M", "top_level_package_classic.sub.bar@{offset}".format(offset=offset)],
        offset=offset,
        expected_sources=[
            "top_level_package_classic/",
            "top_level_package_classic/__init__.py",
            "top_level_package_classic/sub/",
            "top_level_package_classic/sub/__init__.py",
            "top_level_package_classic/sub/bar.py",
        ],
    )


def test_overlap(source_tree):
    # type: (SourceTree) -> None
    source_tree.assert_sources(
        pex_args=[
            "-M",
            "top_level_package_classic.sub.foo@src",
            "-P",
            "top_level_package_classic.sub@src",
        ],
        offset="src",
        expected_sources=[
            "top_level_package_classic/",
            "top_level_package_classic/__init__.py",
            "top_level_package_classic/sub/",
            "top_level_package_classic/sub/__init__.py",
            "top_level_package_classic/sub/foo.py",
            "top_level_package_classic/sub/bar.py",
        ],
    )



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from collections import defaultdict

from pex import pex_warnings
from pex.common import pluralize
from pex.dependency_configuration import DependencyConfiguration
from pex.dist_metadata import Requirement
from pex.environment import PEXEnvironment
from pex.exceptions import production_assert
from pex.fingerprinted_distribution import FingerprintedDistribution
from pex.orderedset import OrderedSet
from pex.pep_503 import ProjectName
from pex.pex_builder import PEXBuilder
from pex.pex_info import PexInfo
from pex.resolve.resolvers import ResolveResult
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import DefaultDict

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s
class DependencyManager(object):
    _requirements = attr.ib(factory=OrderedSet)  # type: OrderedSet[Requirement]
    _distributions = attr.ib(factory=OrderedSet)  # type: OrderedSet[FingerprintedDistribution]

    def add_requirement(self, requirement):
        # type: (Requirement) -> None
        self._requirements.add(requirement)

    def add_distribution(self, fingerprinted_distribution):
        # type: (FingerprintedDistribution) -> None
        self._distributions.add(fingerprinted_distribution)

    def add_from_pex(
        self,
        pex,  # type: str
        result_type_wheel_file=False,  # type: bool
    ):
        # type: (...) -> PexInfo

        pex_info = PexInfo.from_pex(pex)
        for req in pex_info.requirements:
            self.add_requirement(Requirement.parse(req))

        pex_environment = PEXEnvironment.mount(pex, pex_info=pex_info)
        for dist in pex_environment.iter_distributions(
            result_type_wheel_file=result_type_wheel_file
        ):
            self.add_distribution(dist)

        return pex_info

    def add_from_resolved(self, resolved):
        # type: (ResolveResult) -> None

        for resolved_dist in resolved.distributions:
            for req in resolved_dist.direct_requirements:
                self.add_requirement(req)
            self.add_distribution(resolved_dist.fingerprinted_distribution)

    def configure(
        self,
        pex_builder,  # type: PEXBuilder
        dependency_configuration=DependencyConfiguration(),  # type: DependencyConfiguration
    ):
        # type: (...) -> None

        dependency_configuration.configure(pex_builder.info)

        root_requirements_by_project_name = defaultdict(
            OrderedSet
        )  # type: DefaultDict[ProjectName, OrderedSet[Requirement]]
        for root_req in self._requirements:
            root_requirements_by_project_name[root_req.project_name].add(root_req)

        for fingerprinted_dist in self._distributions:
            excluded_by = dependency_configuration.excluded_by(fingerprinted_dist.distribution)
            if excluded_by:
                excludes = " and ".join(map(str, excluded_by))
                root_reqs = root_requirements_by_project_name[fingerprinted_dist.project_name]
                production_assert(
                    len(root_reqs) > 0,
                    "The deep --exclude mechanism failed to exclude {dist} from transitive "
                    "requirements. It should have been excluded by configured excludes: "
                    "{excludes} but was not.".format(
                        dist=fingerprinted_dist.distribution, excludes=excludes
                    ),
                )
                pex_warnings.warn(
                    "The distribution {dist} was required by the input {requirements} "
                    "{root_reqs} but ultimately excluded by configured excludes: "
                    "{excludes}".format(
                        dist=fingerprinted_dist.distribution,
                        requirements=pluralize(root_reqs, "requirement"),
                        root_reqs=" and ".join(map(str, root_reqs)),
                        excludes=excludes,
                    )
                )
                continue
            pex_builder.add_distribution(
                dist=fingerprinted_dist.distribution, fingerprint=fingerprinted_dist.fingerprint
            )

        for requirement in self._requirements:
            pex_builder.add_requirement(requirement)



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import logging
from argparse import Namespace, _ActionsContainer
from textwrap import dedent

from pex import docs
from pex.commands.command import try_open_file
from pex.docs.server import SERVER_NAME, LaunchError, LaunchResult
from pex.docs.server import launch as launch_docs_server
from pex.result import Error, try_
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional, Union

    import attr  # vendor:skip
else:
    from pex.third_party import attr


logger = logging.getLogger(__name__)


def register_open_options(parser):
    # type: (_ActionsContainer) -> None
    parser.add_argument(
        "--browser",
        dest="browser",
        default=None,
        help="The browser to use to open docs with. Defaults to the system default opener.",
    )


@attr.s(frozen=True)
class HtmlDocsConfig(object):
    @classmethod
    def from_options(
        cls,
        options,  # type: Namespace
        fallback_url=None,  # type: Optional[str]
    ):
        # type: (...) -> HtmlDocsConfig
        return cls(browser=options.browser, fallback_url=fallback_url)

    browser = attr.ib(default=None)  # type: Optional[str]
    fallback_url = attr.ib(default=None)  # type: Optional[str]


# PEX in ascii-hex rotated left 1 character and trailing 0 dropped:
# E    X    P
# 0x45 0x58 0x50
#
# This is a currently unassigned port in the 1024-49151 user port registration range.
# This value plays by the loose rules and satisfies the criteria of being unlikely to be in use.
# We have no intention of registering with IANA tough!
STANDARD_PORT = 45585


def serve_html_docs(
    open_browser=False,  # type: bool
    config=HtmlDocsConfig(),  # type: HtmlDocsConfig
):
    # type: (...) -> Union[LaunchResult, Error]
    html_docs = docs.root(doc_type="html")
    if not html_docs:
        return Error(
            dedent(
                """\
                This Pex distribution does not include embedded docs.

                You can find the latest docs here:
                HTML: https://docs.pex-tool.org
                 PDF: https://github.com/pex-tool/pex/releases/latest/download/pex.pdf
                """
            ).rstrip()
        )

    try:
        result = launch_docs_server(html_docs, port=STANDARD_PORT)
    except LaunchError:
        try:
            result = launch_docs_server(html_docs, port=0)
        except LaunchError as e:
            with open(e.log) as fp:
                for line in fp:
                    logger.log(logging.ERROR, line.rstrip())
            return Error("Failed to launch {server}.".format(server=SERVER_NAME))

    if open_browser:
        try_(
            try_open_file(result.server_info.url, open_program=config.browser, suppress_stderr=True)
        )

    return result



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os.path
import re
import shutil
import subprocess
from collections import OrderedDict
from subprocess import CalledProcessError

from pex.atomic_directory import atomic_directory
from pex.common import chmod_plus_x, is_exe, pluralize, safe_mkdtemp, safe_open
from pex.compatibility import shlex_quote
from pex.exceptions import production_assert
from pex.fetcher import URLFetcher
from pex.hashing import Sha256
from pex.layout import Layout
from pex.pep_440 import Version
from pex.pex_info import PexInfo
from pex.result import Error, try_
from pex.scie.model import ScieConfiguration, ScieInfo, SciePlatform, ScieStyle, ScieTarget
from pex.third_party.packaging.specifiers import SpecifierSet
from pex.third_party.packaging.version import InvalidVersion
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING, cast
from pex.util import CacheHelper
from pex.variables import ENV, Variables, unzip_dir_relpath

if TYPE_CHECKING:
    from typing import Any, Dict, Iterator, Optional, Union, cast

    import attr  # vendor:skip
    import toml  # vendor:skip
else:
    from pex.third_party import attr, toml


@attr.s(frozen=True)
class Manifest(object):
    target = attr.ib()  # type: ScieTarget
    path = attr.ib()  # type: str

    def binary_name(self, binary_name):
        # type: (str) -> str
        return self.target.platform.binary_name(binary_name)

    def qualified_binary_name(self, binary_name):
        # type: (str) -> str
        return self.target.platform.qualified_binary_name(binary_name)


SCIENCE_RELEASES_URL = "https://github.com/a-scie/lift/releases"
MIN_SCIENCE_VERSION = Version("0.3.0")
SCIENCE_REQUIREMENT = SpecifierSet("~={min_version}".format(min_version=MIN_SCIENCE_VERSION))


def _science_binary_url(suffix=""):
    # type: (str) -> str
    return "{science_releases_url}/download/v{version}/{binary}{suffix}".format(
        science_releases_url=SCIENCE_RELEASES_URL,
        version=MIN_SCIENCE_VERSION.raw,
        binary=SciePlatform.CURRENT.qualified_binary_name("science-fat"),
        suffix=suffix,
    )


PTEX_VERSION = "1.1.1"
SCIE_JUMP_VERSION = "1.1.1"


def create_manifests(
    configuration,  # type: ScieConfiguration
    name,  # type: str
    pex_info,  # type: PexInfo
    layout,  # type: Layout.Value
):
    # type: (...) -> Iterator[Manifest]

    pex_root = "{scie.bindings}/pex_root"
    if pex_info.venv:
        # We let the configure-binding calculate the venv dir at runtime since it depends on the
        # interpreter executing the venv PEX.
        installed_pex_dir = ""
    elif layout is Layout.LOOSE:
        installed_pex_dir = "{pex}"
    else:
        production_assert(pex_info.pex_hash is not None)
        pex_hash = cast(str, pex_info.pex_hash)
        installed_pex_dir = os.path.join(pex_root, unzip_dir_relpath(pex_hash))

    env_default = {
        "PEX_ROOT": pex_root,
    }

    lift = {
        "name": name,
        "ptex": {
            "id": "ptex",
            "version": PTEX_VERSION,
            "argv1": "{scie.env.PEX_BOOTSTRAP_URLS={scie.lift}}",
        },
        "scie_jump": {"version": SCIE_JUMP_VERSION},
        "files": [{"name": "configure-binding.py"}, {"name": "pex"}],
        "commands": [
            {
                "env": {"default": env_default},
                "exe": "{scie.bindings.configure:PYTHON}",
                "args": ["{scie.bindings.configure:PEX}"],
            }
        ],
        "bindings": [
            {
                "env": {
                    "default": env_default,
                    "remove_exact": ["PATH"],
                    "remove_re": ["PEX_.*"],
                    "replace": {
                        "PEX_INTERPRETER": "1",
                        "_PEX_SCIE_INSTALLED_PEX_DIR": installed_pex_dir,
                        # We can get a warning about too-long script shebangs, but this is not
                        # relevant since we above run the PEX via python and not via shebang.
                        "PEX_EMIT_WARNINGS": "0",
                    },
                },
                "name": "configure",
                "exe": "#{cpython:python}",
                "args": ["{pex}", "{configure-binding.py}"],
            }
        ],
    }  # type: Dict[str, Any]

    for target in configuration.targets:
        manifest_path = os.path.join(
            safe_mkdtemp(),
            target.platform.qualified_file_name("{name}-lift.toml".format(name=name)),
        )
        with safe_open(manifest_path, "w") as fp:
            toml.dump(
                {
                    "lift": dict(
                        lift,
                        platforms=[target.platform.value],
                        interpreters=[
                            {
                                "id": "cpython",
                                "provider": "PythonBuildStandalone",
                                "release": target.pbs_release,
                                "version": target.version_str,
                                "lazy": configuration.options.style is ScieStyle.LAZY,
                            }
                        ],
                    )
                },
                fp,
            )
        yield Manifest(target=target, path=manifest_path)


def _science_dir(
    env,  # type: Variables
    *components  # type: str
):
    # type: (...) -> str
    return os.path.join(env.PEX_ROOT, "scies", "science", MIN_SCIENCE_VERSION.raw, *components)


def _science_binary_names():
    # type: () -> Iterator[str]
    yield SciePlatform.CURRENT.binary_name("science-fat")
    yield SciePlatform.CURRENT.qualified_binary_name("science-fat")
    yield SciePlatform.CURRENT.binary_name("science")
    yield SciePlatform.CURRENT.qualified_binary_name("science")


def _is_compatible_science_binary(
    binary,  # type: str
    source=None,  # type: Optional[str]
):
    # type: (...) -> Union[Version, Error]
    try:
        version = Version(
            subprocess.check_output(args=[binary, "--version"]).decode("utf-8").strip()
        )
    except (CalledProcessError, InvalidVersion) as e:
        return Error(
            "Failed to determine --version of science binary at {source}: {err}".format(
                source=source or binary, err=e
            )
        )
    else:
        if version.raw in SCIENCE_REQUIREMENT:
            return version
        return Error(
            "The science binary at {source} is version {version} which does not match Pex's "
            "science requirement of {science_requirement}.".format(
                source=source or binary,
                version=version.raw,
                science_requirement=SCIENCE_REQUIREMENT,
            )
        )


def _path_science():
    # type: () -> Optional[str]
    for path_element in os.environ.get("PATH", os.defpath).split(os.pathsep):
        for binary in (
            os.path.join(path_element, binary_name) for binary_name in _science_binary_names()
        ):
            if not is_exe(binary):
                continue
            if isinstance(_is_compatible_science_binary(binary), Error):
                continue
            return binary
    return None


def _ensure_science(
    url_fetcher=None,  # type: Optional[URLFetcher]
    science_binary_url=None,  # type: Optional[str]
    env=ENV,  # type: Variables
):
    # type: (...) -> str

    target_dir = _science_dir(env, "bin")
    with atomic_directory(target_dir=target_dir) as atomic_dir:
        if not atomic_dir.is_finalized():
            target_science = os.path.join(atomic_dir.work_dir, "science")
            path_science = _path_science()
            if path_science:
                shutil.copy(path_science, target_science)
            else:
                fetcher = url_fetcher or URLFetcher()
                with open(target_science, "wb") as write_fp, fetcher.get_body_stream(
                    science_binary_url or _science_binary_url()
                ) as read_fp:
                    shutil.copyfileobj(read_fp, write_fp)
                chmod_plus_x(target_science)

                if science_binary_url:
                    custom_science_binary_version = try_(
                        _is_compatible_science_binary(target_science, source=science_binary_url)
                    )
                    TRACER.log(
                        "Using custom science binary from {source} with version {version}.".format(
                            source=science_binary_url, version=custom_science_binary_version.raw
                        )
                    )
                else:
                    # Since we used the canonical GitHub Releases URL, we know a checksum file is
                    # available we can use to verify.
                    science_sha256_url = _science_binary_url(".sha256")
                    with fetcher.get_body_stream(science_sha256_url) as fp:
                        expected_sha256, _, _ = fp.read().decode("utf-8").partition(" ")
                    actual_sha256 = CacheHelper.hash(target_science, hasher=Sha256)
                    if expected_sha256 != actual_sha256:
                        raise ValueError(
                            "The science binary downloaded from {science_binary_url} does not "
                            "match the expected SHA-256 fingerprint recorded in "
                            "{science_sha256_url}.\n"
                            "Expected {expected_sha256} but found {actual_sha256}.".format(
                                science_binary_url=science_binary_url,
                                science_sha256_url=science_sha256_url,
                                expected_sha256=expected_sha256,
                                actual_sha256=actual_sha256,
                            )
                        )
    return os.path.join(target_dir, "science")


class ScienceError(Exception):
    """Indicates an error executing science."""


def build(
    configuration,  # type: ScieConfiguration
    pex_file,  # type: str
    url_fetcher=None,  # type: Optional[URLFetcher]
    env=ENV,  # type: Variables
):
    # type: (...) -> Iterator[ScieInfo]

    science = _ensure_science(
        url_fetcher=url_fetcher,
        science_binary_url=configuration.options.science_binary_url,
        env=env,
    )
    name = re.sub(r"\.pex$", "", os.path.basename(pex_file), flags=re.IGNORECASE)
    pex_info = PexInfo.from_pex(pex_file)
    layout = Layout.identify(pex_file)
    use_platform_suffix = len(configuration.targets) > 1

    errors = OrderedDict()  # type: OrderedDict[Manifest, str]
    for manifest in create_manifests(configuration, name, pex_info, layout):
        args = [science, "--cache-dir", _science_dir(env, "cache")]
        if env.PEX_VERBOSE:
            args.append("-{verbosity}".format(verbosity="v" * env.PEX_VERBOSE))
        dest_dir = os.path.dirname(os.path.abspath(pex_file))
        args.extend(
            [
                "lift",
                "--file",
                "pex={pex_file}".format(pex_file=pex_file),
                "--file",
                "configure-binding.py={configure_binding}".format(
                    configure_binding=os.path.join(
                        os.path.dirname(__file__), "configure-binding.py"
                    )
                ),
                "build",
                "--dest-dir",
                dest_dir,
            ]
        )
        if use_platform_suffix:
            args.append("--use-platform-suffix")
        args.append(manifest.path)
        with open(os.devnull, "wb") as devnull:
            process = subprocess.Popen(args=args, stdout=devnull, stderr=subprocess.PIPE)
            _, stderr = process.communicate()
            if process.returncode != 0:
                saved_manifest = os.path.relpath(
                    os.path.join(dest_dir, os.path.basename(manifest.path))
                )
                shutil.copy(manifest.path, saved_manifest)
                errors[manifest] = (
                    "Command `{command}` failed with exit code {exit_code} (saved lift manifest to "
                    "{saved_manifest} for inspection):\n{stderr}"
                ).format(
                    command=" ".join(shlex_quote(arg) for arg in args[:-1] + [saved_manifest]),
                    exit_code=process.returncode,
                    saved_manifest=saved_manifest,
                    stderr=stderr.decode("utf-8").strip(),
                )
            else:
                yield ScieInfo(
                    style=configuration.options.style,
                    target=manifest.target,
                    file=os.path.join(
                        dest_dir,
                        manifest.qualified_binary_name(name)
                        if use_platform_suffix
                        else manifest.binary_name(name),
                    ),
                )

    if errors:
        raise ScienceError(
            "Failed to build {count} {scies}:\n\n{errors}".format(
                count=len(errors),
                scies=pluralize(errors, "scie"),
                errors="\n\n".join(
                    "{index}. For CPython {version} on {platform}: {err}".format(
                        index=index,
                        platform=manifest.target.platform,
                        version=manifest.target.version_str,
                        err=err,
                    )
                    for index, (manifest, err) in enumerate(errors.items(), start=1)
                ),
            )
        )



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import json
from textwrap import dedent

from testing.docker import DockerVirtualenvRunner


def test_installed_wheels(fedora39_virtualenv_runner):
    # type: (DockerVirtualenvRunner) -> None

    assert {
        "ansicolors==1.1.8": "/virtualenv.venv/lib/python3.12/site-packages",
        "numpy==1.26.4": "/virtualenv.venv/lib64/python3.12/site-packages",
    } == json.loads(
        fedora39_virtualenv_runner.run(
            dedent(
                """\
                import json
                import subprocess
                import sys

                from pex.venv.virtualenv import Virtualenv


                subprocess.check_call(
                    args=[
                        sys.executable,
                        "-m",
                        "pex.cli",
                        "venv",
                        "create",
                        "-d",
                        "/virtualenv.venv",
                        "ansicolors==1.1.8",
                        "numpy==1.26.4"
                    ],
                    stdout=sys.stderr
                )

                venv = Virtualenv("/virtualenv.venv")
                json.dump(
                    {
                        str(dist.as_requirement()): dist.location
                        for dist in venv.iter_distributions()
                        if str(dist.metadata.project_name) in ("ansicolors", "numpy")
                    },
                    sys.stdout
                )
                """
            )
        ).decode("utf-8")
    )


def test_wheel_files(fedora39_virtualenv_runner):
    # type: (DockerVirtualenvRunner) -> None

    assert {
        "ansicolors==1.1.7": "/virtualenv.venv/lib/python3.12/site-packages",
        "numpy==1.26.3": "/virtualenv.venv/lib64/python3.12/site-packages",
    } == json.loads(
        fedora39_virtualenv_runner.run(
            dedent(
                """\
                import glob
                import json
                import subprocess
                import sys

                from pex.interpreter import PythonInterpreter
                from pex.pep_427 import install_wheel_interpreter
                from pex.venv.virtualenv import Virtualenv


                subprocess.check_call(
                    args=[
                        sys.executable,
                        "-m",
                        "pip",
                        "wheel",
                        "-w",
                        "/wheels",
                        "ansicolors==1.1.7",
                        "numpy==1.26.3"
                    ],
                    stdout=sys.stderr
                )

                interpreter = PythonInterpreter.get()
                for wheel_path in glob.glob("/wheels/*.whl"):
                    install_wheel_interpreter(wheel_path=wheel_path, interpreter=interpreter)

                venv = Virtualenv("/virtualenv.venv")
                json.dump(
                    {
                        str(dist.as_requirement()): dist.location
                        for dist in venv.iter_distributions()
                        if str(dist.metadata.project_name) in ("ansicolors", "numpy")
                    },
                    sys.stdout
                )
                """
            )
        ).decode("utf-8")
    )



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os.path
import re
import shutil
import sys
from collections import defaultdict

import pytest

from pex.dist_metadata import ProjectNameAndVersion, Requirement
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.pex import PEX
from pex.pex_info import PexInfo
from pex.typing import TYPE_CHECKING
from testing import PY39, PY310, PY_VER, data, ensure_python_interpreter, run_pex_command
from testing.cli import run_pex3
from testing.lock import extract_lock_option_args, index_lock_artifacts

if TYPE_CHECKING:
    from typing import Any, Iterable, List, Mapping


def assert_overrides(
    pex,  # type: str
    expected_overrides,  # type: Iterable[str]
    expected_overridden_dists,  # type: Mapping[str, List[str]]
):
    pex_info = PexInfo.from_pex(pex)
    assert list(sorted(expected_overrides)) == list(sorted(pex_info.overridden))

    dists_to_versions = defaultdict(list)
    for dist in pex_info.distributions:
        project_name_and_version = ProjectNameAndVersion.from_filename(dist)
        dists_to_versions[project_name_and_version.canonicalized_project_name].append(
            project_name_and_version.canonicalized_version
        )

    expected_project_names_to_versions = {
        ProjectName(project_name): [Version(version) for version in sorted(versions)]
        for project_name, versions in expected_overridden_dists.items()
    }
    assert expected_project_names_to_versions == {
        project_name: sorted(versions, key=str)
        for project_name, versions in dists_to_versions.items()
        if project_name in expected_project_names_to_versions
    }, pex_info.dump(indent=2)


skip_unless_compatible_with_requests_2_31_0 = pytest.mark.skipif(
    PY_VER < (3, 7), reason="The requests version tested requires Python `>=3.7`."
)


@skip_unless_compatible_with_requests_2_31_0
def test_override(tmpdir):
    # type: (Any) -> None

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(
        args=[
            "--python",
            sys.executable,
            "--python",
            ensure_python_interpreter(PY310 if PY_VER < (3, 10) else PY39),
            "requests==2.31.0",
            "--override",
            "urllib3==1.21; python_version >= '3.10'",
            "--override",
            "urllib3==1.20; python_version < '3.10'",
            "--override",
            "idna<2.5",
            "-o",
            pex,
        ]
    ).assert_success()

    dists = [
        dist for dist in PEX(pex).resolve() if ProjectName("requests") == dist.metadata.project_name
    ]
    assert 1 == len(dists)
    requests = dists[0]
    request_deps_by_project_name = {dep.project_name: dep for dep in requests.requires()}
    assert (
        Requirement.parse("urllib3<3,>=1.21.1")
        == request_deps_by_project_name[ProjectName("urllib3")]
    ), (
        "Our multiple override test requires requests normally have an urllib3 lower bound of "
        "1.21.1."
    )
    assert (
        Requirement.parse("idna<4,>=2.5") == request_deps_by_project_name[ProjectName("idna")]
    ), "Our single override test requires requests normally have an idna lower bound of 2.5."

    assert_overrides(
        pex,
        expected_overrides=[
            'urllib3==1.21; python_version >= "3.10"',
            'urllib3==1.20; python_version < "3.10"',
            "idna<2.5",
        ],
        expected_overridden_dists={"urllib3": ["1.20", "1.21"], "idna": ["2.4"]},
    )


@skip_unless_compatible_with_requests_2_31_0
def test_pex_repository_override(tmpdir):
    # type: (Any) -> None

    repository_pex = os.path.join(str(tmpdir), "repository.pex")
    run_pex_command(
        args=["requests==2.31.0", "--override", "idna<2.4", "-o", repository_pex]
    ).assert_success()

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(
        args=["--pex-repository", repository_pex, "requests", "-o", pex]
    ).assert_success()
    assert_overrides(
        pex, expected_overrides=["idna<2.4"], expected_overridden_dists={"idna": ["2.3"]}
    )


REQUESTS_LOCK = data.path("locks", "requests.lock.json")


@pytest.fixture
def requests_lock(tmpdir):
    # type: (Any) -> str

    lock = os.path.join(str(tmpdir), "requests-lock.json")
    shutil.copy(REQUESTS_LOCK, lock)
    return lock


skip_unless_compatible_with_requests_lock = pytest.mark.skipif(
    PY_VER < (3, 7) or PY_VER >= (3, 13), reason="The lock used is for >=3.7,<3.13"
)


@skip_unless_compatible_with_requests_lock
def test_lock_sync_override(
    tmpdir,  # type: Any
    requests_lock,  # type: str
):
    # type: (...) -> None

    index = index_lock_artifacts(requests_lock)
    locked_requests = index[ProjectName("requests")]
    request_deps_by_project_name = defaultdict(list)
    for req in locked_requests.requires_dists:
        request_deps_by_project_name[req.project_name].append(req)
    assert [Requirement.parse("urllib3<3,>=1.21.1")] == request_deps_by_project_name[
        ProjectName("urllib3")
    ], (
        "We expect the locked requests to impose a lower bound of 1.21.1 on urllib3 since we'll "
        "try to edit that lower bound below using `--override`."
    )

    result = run_pex3(
        *(
            ["lock", "sync", "--lock", requests_lock, "--override", "urllib3<1.21.1"]
            + extract_lock_option_args(requests_lock)
        )
    )
    result.assert_success()
    assert "Updates for lock generated by universal:\n" in result.error, result.error
    assert "  Updated urllib3 from 2.0.7 to 1.21\n" in result.error, result.error

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(args=["--lock", requests_lock, "-o", pex]).assert_success()
    assert_overrides(
        pex, expected_overrides=["urllib3<1.21.1"], expected_overridden_dists={"urllib3": ["1.21"]}
    )


@skip_unless_compatible_with_requests_lock
def test_illegal_override(
    tmpdir,  # type: Any
    requests_lock,  # type: str
):
    # type: (...) -> None

    pex = os.path.join(str(tmpdir), "pex")
    run_pex_command(args=["--lock", requests_lock, "--override", "foo", "-o", pex]).assert_failure(
        expected_error_re=".*{expected_message}$".format(
            expected_message=re.escape(
                "The --override option cannot be used when resolving against a lock file. "
                "Only overrides already present in the lock file will be applied.\n"
            )
        ),
        re_flags=re.DOTALL,
    )

    pex_repository = os.path.join(str(tmpdir), "repository.pex")
    run_pex_command(args=["--lock", requests_lock, "-o", pex_repository]).assert_success()
    run_pex_command(
        args=["--pex-repository", requests_lock, "--override", "bar", "-o", pex]
    ).assert_failure(
        expected_error_re=".*{expected_message}$".format(
            expected_message=re.escape(
                "The --override option cannot be used when resolving against a PEX repository. "
                "Only overrides already present in the PEX repository will be applied.\n"
            )
        ),
        re_flags=re.DOTALL,
    )



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import re
from textwrap import dedent

import pytest

from pex.dist_metadata import Requirement
from pex.pep_723 import InvalidMetadataError, ScriptMetadata
from pex.third_party.packaging import specifiers
from testing import PY_VER


def test_parse_empty():
    # type: () -> None

    assert not ScriptMetadata.parse("")
    assert not ScriptMetadata.parse(
        dedent(
            """\
            # /// script
            # ///
            """
        )
    )
    assert not ScriptMetadata.parse(
        dedent(
            """\
            # /// script
            # unknown_key = 42
            # ///
            """
        )
    )
    assert not ScriptMetadata.parse(
        dedent(
            """\
            # /// script
            # unknown_key = 42
            # ///
            # dependencies = ["ansicolors"]
            """
        )
    )


def test_parse_unterminated():
    # type: () -> None

    assert not ScriptMetadata.parse(
        dedent(
            """\
            # /// script
            # dependencies = ["ansicolors"]

            # # N.B.: The line above does not start with # which should terminate the search for
            # # the terminator immediately below.
            # ///
            """
        )
    )
    assert not ScriptMetadata.parse(
        dedent(
            """\
            # /// script
            # dependencies = ["ansicolors"]
            """
        )
    )


def test_parse_greedy():
    # type: () -> None

    with pytest.raises(
        InvalidMetadataError,
        match=re.escape(
            "Invalid specifier: 'Invalid specifier with embedded script metadata block trojan.\n"
            "///'"
        ),
    ):
        ScriptMetadata.parse(
            dedent(
                '''\
                # /// script
                # requires-python = """
                # Invalid specifier with embedded script metadata block trojan.
                # ///
                # """
                # ///
                '''
            )
        )


def test_parse_nominal():
    # type: () -> None

    assert ScriptMetadata(
        dependencies=tuple([Requirement.parse("ansicolors")])
    ) == ScriptMetadata.parse(
        dedent(
            """
            # /// script
            # dependencies = ["ansicolors"]
            # ///
            """
        )
    )

    assert ScriptMetadata(requires_python=specifiers.SpecifierSet("~=3.8")) == ScriptMetadata.parse(
        dedent(
            """
            # /// script
            # requires-python = "~=3.8"
            # ///
            """
        )
    )

    assert ScriptMetadata(
        dependencies=tuple([Requirement.parse("cowsay<6")]),
        requires_python=specifiers.SpecifierSet("==2.7.*"),
    ) == ScriptMetadata.parse(
        dedent(
            """
            dependencies = ["before"]
            # /// script
            # dependencies = [
            #   "cowsay<6",
            # ]
            #
            # # Yup, 2.7.
            # requires-python = "==2.7.*"
            #
            # not-a-recognized-key = 42
            # ///
            dependencies = ["after"]
            """
        )
    )


def test_parse_invalid_embedded_start():
    # type: () -> None

    with pytest.raises(
        InvalidMetadataError,
        match=re.escape(
            "The script metadata found in <unspecified source> contains a `# /// script` block "
            "beginning on line 1 that is followed by a `# /// script` block beginning on line 2 "
            "before the 'script' block starting on line 1 is closed.\n"
            "Metadata blocks must be closed before a new metadata block can begin.\n"
            "See: https://packaging.python.org/specifications/"
            "inline-script-metadata#inline-script-metadata"
        ),
    ):
        ScriptMetadata.parse(
            dedent(
                """\
                # /// script
                # /// script
                # ///
                """
            )
        )

    with pytest.raises(
        InvalidMetadataError,
        match=re.escape(
            "The script metadata found in <unspecified source> contains a `# /// script` block "
            "beginning on line 3 that is followed by a `# /// future` block beginning on line 5 "
            "before the 'script' block starting on line 3 is closed.\n"
            "Metadata blocks must be closed before a new metadata block can begin.\n"
            "See: https://packaging.python.org/specifications/"
            "inline-script-metadata#inline-script-metadata"
        ),
    ):
        ScriptMetadata.parse(
            dedent(
                """\
                # Preamble.
                #
                # /// script
                #
                # /// future
                # ///
                """
            )
        )


def test_parse_invalid_multiple_blocks():
    # type: () -> None

    with pytest.raises(
        InvalidMetadataError,
        match=re.escape(
            "Found 2 metadata block types in <unspecified source> with more than one appearance:\n"
            "+ 2 `# /// script` metadata blocks beginning on lines 3 and 8.\n"
            "+ 3 `# /// future` metadata blocks beginning on lines 6, 11 and 15.\n"
            "At most one metadata block of each type is allowed.\n"
            "See: https://packaging.python.org/specifications/"
            "inline-script-metadata#inline-script-metadata"
        ),
    ):
        ScriptMetadata.parse(
            dedent(
                """\
                # A script that uses PEP-723, badly.
                #
                # /// script
                # ///

                # /// future
                # ///
                # /// script
                # ///
                #
                # /// future
                # ///
                # /// bob
                # ///
                # /// future
                # ///
                """
            )
        )


def test_parse_invalid_toml():
    # type: () -> None

    with pytest.raises(
        InvalidMetadataError,
        match=re.escape(
            "The script metadata found in exe.py starting at line 3 embeds malformed toml: "
            "Empty value is invalid (line 1 column 1 char 0).\n"
            "See: https://packaging.python.org/specifications/"
            "inline-script-metadata#inline-script-metadata"
        ),
    ):
        ScriptMetadata.parse(
            dedent(
                """\
                # Preamble.
                #
                # /// script
                # unspecified_value = # These are invalid in TOML; you must specify a value.
                # ///
                """
            ),
            source="exe.py",
        )


def test_parse_invalid_dependencies_value_type():
    # type: () -> None

    with pytest.raises(
        InvalidMetadataError,
        match=re.escape(
            "The script metadata found in <unspecified source> starting at line 1 contains an "
            "invalid `dependencies` value of type `int`\n"
            "Expected a list of dependency specifier strings.\n"
            "See: https://packaging.python.org/specifications/"
            "inline-script-metadata#inline-script-metadata"
        ),
    ):
        ScriptMetadata.parse(
            dedent(
                """\
                # /// script
                # dependencies = 42
                # ///
                """
            )
        )


@pytest.mark.skipif(
    PY_VER < (3, 7),
    reason=(
        "The version of vendored packaging used for Python>=3.7 is required for the precise error "
        "message being tested against."
    ),
)
def test_parse_invalid_dependencies_values():
    # type: () -> None

    with pytest.raises(
        InvalidMetadataError,
        match=re.escape(
            "The script metadata found in <unspecified source> starting at line 1 contains a "
            "`dependencies` list with 2 invalid dependency specifiers:\n"
            "+ dependencies[1] '; python_version >= \"3.8\"': Expected package name at the "
            "start of dependency specifier\n"
            '    ; python_version >= "3.8"\n'
            "    ^\n"
            "+ dependencies[3] '|': Expected package name at the start of dependency specifier\n"
            "    |\n"
            "    ^.\n"
            "See: https://packaging.python.org/specifications/"
            "inline-script-metadata#inline-script-metadata"
        ),
    ):
        ScriptMetadata.parse(
            dedent(
                """\
                # /// script
                # dependencies = [
                #     "ansicolors",
                #     "; python_version >= \\"3.8\\"",
                #     "cowsay",
                #     "|",
                # ]
                # ///
                """
            )
        )


def test_parse_invalid_requires_python_value_type():
    # type: () -> None

    with pytest.raises(
        InvalidMetadataError,
        match=re.escape(
            "The script metadata found in <unspecified source> starting at line 1 contains an "
            "invalid `requires-python` value of type `int`\n"
            "Expected a version specifier string.\n"
            "See: https://packaging.python.org/specifications/"
            "inline-script-metadata#inline-script-metadata"
        ),
    ):
        ScriptMetadata.parse(
            dedent(
                """\
                # /// script
                # requires-python = 42
                # ///
                """
            )
        )


@pytest.mark.skipif(
    PY_VER < (3, 7),
    reason=(
        "The version of vendored packaging used for Python>=3.7 is required for the precise error "
        "message being tested against."
    ),
)
def test_parse_invalid_requires_python_value():
    # type: () -> None

    with pytest.raises(
        InvalidMetadataError,
        match=re.escape(
            "The script metadata found in <unspecified source> starting at line 1 contains an "
            "invalid `requires-python` value '>=3.8.*': Invalid specifier: '>=3.8.*'.\n"
            "See: https://packaging.python.org/specifications/"
            "inline-script-metadata#inline-script-metadata"
        ),
    ):
        ScriptMetadata.parse(
            dedent(
                """\
                # /// script
                # requires-python = ">=3.8.*"
                # ///
                """
            )
        )



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path

import pytest

from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.pip.log_analyzer import ErrorMessage
from pex.requirements import parse_requirement_string
from pex.resolve.configured_resolver import ConfiguredResolver
from pex.resolve.locked_resolve import LockConfiguration, LockStyle
from pex.resolve.locker import Locker, LockResult
from pex.resolve.resolved_requirement import (
    ArtifactURL,
    Fingerprint,
    PartialArtifact,
    Pin,
    ResolvedRequirement,
)
from pex.targets import LocalInterpreter
from pex.typing import TYPE_CHECKING
from testing import data

if TYPE_CHECKING:
    from typing import Any


@pytest.fixture
def locker(tmpdir):
    # type: (Any) -> Locker
    download_dir = os.path.join(str(tmpdir), "downloads")
    return Locker(
        target=LocalInterpreter.create(),
        root_requirements=[parse_requirement_string("wheel")],
        resolver=ConfiguredResolver.default(),
        lock_configuration=LockConfiguration(style=LockStyle.SOURCES),
        download_dir=download_dir,
    )


def analyze_log(
    locker,  # type: Locker
    log_name,  # type: str
):
    # type: (...) -> None
    with open(data.path("pip_logs", log_name)) as fp:
        for line in fp:
            result = locker.analyze(line)
            assert not isinstance(result.data, ErrorMessage)
    locker.analysis_completed()


@pytest.mark.parametrize("pip_version", ["23.2", "23.3.1"])
def test_redirects_dont_stomp_original_index_urls(
    pip_version,  # type: str
    locker,  # type: Locker
):
    # type: (...) -> None

    analyze_log(locker, "issue-2414.pip-{pip_version}.log".format(pip_version=pip_version))

    expected_whl_artifact = PartialArtifact(
        url=ArtifactURL.parse(
            "https://m.devpi.net/root/pypi/%2Bf/55c/570405f142630/wheel-0.43.0-py3-none-any.whl"
        ),
        fingerprint=Fingerprint(
            algorithm="sha256",
            hash="55c570405f142630c6b9f72fe09d9b67cf1477fcf543ae5b8dcb1f5b7377da81",
        ),
        verified=False,
    )
    expected_sdist_artifact = PartialArtifact(
        url=ArtifactURL.parse(
            "https://m.devpi.net/root/pypi/%2Bf/465/ef92c69fa5c5d/wheel-0.43.0.tar.gz"
        ),
        fingerprint=Fingerprint(
            algorithm="sha256",
            hash="465ef92c69fa5c5da2d1cf8ac40559a8c940886afcef87dcf14b9470862f1d85",
        ),
        verified=False,
    )
    expected_wheel_requirement = ResolvedRequirement(
        pin=Pin(ProjectName("wheel"), Version("0.43.0")),
        artifact=expected_whl_artifact,
        additional_artifacts=tuple([expected_sdist_artifact]),
    )
    expected_lock_result = LockResult(
        resolved_requirements=tuple([expected_wheel_requirement]),
        local_projects=(),
    )
    assert expected_lock_result == locker.lock_result



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import itertools
import os.path
import re
import shutil
import subprocess
import sys
from contextlib import closing
from fileinput import FileInput
from textwrap import dedent

from pex import pex_warnings
from pex.common import chmod_plus_x, is_pyc_file, iter_copytree, open_zip, safe_open, touch
from pex.compatibility import commonpath, get_stdout_bytes_buffer
from pex.dist_metadata import Distribution, ProjectNameAndVersion
from pex.enum import Enum
from pex.interpreter import PythonInterpreter
from pex.pep_376 import InstalledFile, InstalledWheel, Record
from pex.pep_503 import ProjectName
from pex.typing import TYPE_CHECKING, cast
from pex.wheel import Wheel

if TYPE_CHECKING:
    from typing import (  # noqa
        Callable,
        DefaultDict,
        Iterable,
        Iterator,
        List,
        Optional,
        Text,
        Tuple,
    )

    import attr  # vendor:skip
else:
    from pex.third_party import attr


class WheelError(Exception):
    """Indicates an error interacting with a wheel."""


class InstallableType(Enum["InstallableType.Value"]):
    class Value(Enum.Value):
        pass

    INSTALLED_WHEEL_CHROOT = Value("installed wheel chroot")
    WHEEL_FILE = Value(".whl file")


@attr.s(frozen=True)
class InstallPaths(object):

    CHROOT_STASH = ".prefix"

    @classmethod
    def chroot(
        cls,
        destination,  # type: str
        project_name,  # type: ProjectName
    ):
        # type: (...) -> InstallPaths
        base = os.path.join(destination, cls.CHROOT_STASH)
        return cls(
            purelib=destination,
            platlib=destination,
            headers=os.path.join(base, "include", "site", "pythonX.Y", project_name.raw),
            scripts=os.path.join(base, "bin"),
            data=base,
        )

    @classmethod
    def interpreter(cls, interpreter):
        # type: (PythonInterpreter) -> InstallPaths
        sysconfig_paths = interpreter.identity.paths
        return cls(
            purelib=sysconfig_paths["purelib"],
            platlib=sysconfig_paths["platlib"],
            headers=sysconfig_paths["include"],
            scripts=sysconfig_paths["scripts"],
            data=sysconfig_paths["data"],
        )

    purelib = attr.ib()  # type: str
    platlib = attr.ib()  # type: str
    headers = attr.ib()  # type: str
    scripts = attr.ib()  # type: str
    data = attr.ib()  # type: str

    def __getitem__(self, item):
        # type: (Text) -> str
        if "purelib" == item:
            return self.purelib
        elif "platlib" == item:
            return self.platlib
        elif "headers" == item:
            return self.headers
        elif "scripts" == item:
            return self.scripts
        elif "data" == item:
            return self.data
        raise KeyError("Not a known install path: {item}".format(item=item))


class WheelInstallError(WheelError):
    """Indicates an error installing a `.whl` file."""


def install_wheel_chroot(
    wheel_path,  # type: str
    destination,  # type: str
    compile=False,  # type: bool
    requested=True,  # type: bool
):
    # type: (...) -> InstalledWheel

    wheel = install_wheel(
        wheel_path,
        InstallPaths.chroot(
            destination,
            project_name=ProjectNameAndVersion.from_filename(wheel_path).canonicalized_project_name,
        ),
        compile=compile,
        requested=requested,
    )

    record_relpath = wheel.metadata_files.metadata_file_rel_path("RECORD")
    assert (
        record_relpath is not None
    ), "The {module}.install_wheel function should always create a RECORD.".format(module=__name__)
    return InstalledWheel.save(
        prefix_dir=destination,
        stash_dir=InstallPaths.CHROOT_STASH,
        record_relpath=record_relpath,
        root_is_purelib=wheel.root_is_purelib,
    )


def install_wheel_interpreter(
    wheel_path,  # type: str
    interpreter,  # type: PythonInterpreter
    compile=True,  # type: bool
    requested=True,  # type: bool
):
    # type: (...) -> Wheel

    return install_wheel(
        wheel_path,
        InstallPaths.interpreter(interpreter),
        interpreter=interpreter,
        compile=compile,
        requested=requested,
    )


def install_wheel(
    wheel_path,  # type: str
    install_paths,  # type: InstallPaths
    interpreter=None,  # type: Optional[PythonInterpreter]
    compile=False,  # type: bool
    requested=True,  # type: bool
):
    # type: (...) -> Wheel

    # See: https://packaging.python.org/en/latest/specifications/binary-distribution-format/#installing-a-wheel-distribution-1-0-py32-none-any-whl
    wheel = Wheel.load(wheel_path)
    dest = install_paths.purelib if wheel.root_is_purelib else install_paths.platlib

    record_relpath = wheel.metadata_path("RECORD")
    record_abspath = os.path.join(dest, record_relpath)

    data_rel_path = wheel.data_dir
    data_path = os.path.join(dest, data_rel_path)

    installed_files = []  # type: List[InstalledFile]

    def record_files(
        root_dir,  # type: Text
        names,  # type: Iterable[Text]
    ):
        # type: (...) -> None
        for name in sorted(names):
            if is_pyc_file(name):
                # These files are both optional to RECORD and should never be present in wheels
                # anyway per the spec.
                continue
            file_abspath = os.path.join(root_dir, name)
            if record_relpath == name:
                # We'll generate a new RECORD below.
                os.unlink(file_abspath)
                continue
            installed_files.append(
                InstalledWheel.create_installed_file(path=file_abspath, dest_dir=dest)
            )

    with open_zip(wheel_path) as zf:
        zf.extractall(dest)
        # TODO(John Sirois): Consider verifying signatures.
        # N.B.: Pip does not and its also not clear what good this does. A zip can be easily poked
        # on a per-entry basis allowing forging a RECORD entry and its associated file. Only an
        # outer fingerprint of the whole wheel really solves this sort of tampering.
        record_files(
            root_dir=dest,
            names=[
                name
                for name in zf.namelist()
                if not name.endswith("/") and data_rel_path != commonpath((data_rel_path, name))
            ],
        )
        if os.path.isdir(data_path):
            for entry in sorted(os.listdir(data_path)):
                try:
                    dest_dir = install_paths[entry]
                except KeyError as e:
                    raise WheelInstallError(
                        "The wheel at {wheel_path} is invalid and cannot be installed: "
                        "{err}".format(wheel_path=wheel_path, err=e)
                    )
                entry_path = os.path.join(data_path, entry)
                copied = [dst for _, dst in iter_copytree(entry_path, dest_dir)]
                if copied and "scripts" == entry:
                    for script in copied:
                        chmod_plus_x(script)
                    if interpreter:
                        with closing(FileInput(files=copied, inplace=True, mode="rb")) as script_fi:
                            for line in cast("Iterator[bytes]", script_fi):
                                buffer = get_stdout_bytes_buffer()
                                if script_fi.isfirstline() and re.match(br"^#!pythonw?", line):
                                    _, _, shebang_args = line.partition(b" ")
                                    buffer.write(
                                        "{shebang}\n".format(
                                            shebang=interpreter.shebang(
                                                args=shebang_args.decode("utf-8")
                                            )
                                        ).encode("utf-8")
                                    )
                                else:
                                    # N.B.: These lines include the newline already.
                                    buffer.write(cast(bytes, line))

                record_files(
                    root_dir=dest_dir,
                    names=[
                        os.path.relpath(os.path.join(root, f), entry_path)
                        for root, _, files in os.walk(entry_path)
                        for f in files
                    ],
                )
            shutil.rmtree(data_path)

    if compile:
        args = [
            interpreter.binary if interpreter else sys.executable,
            "-sE",
            "-m",
            "compileall",
        ]  # type: List[Text]
        py_files = [
            os.path.join(dest, installed_file.path)
            for installed_file in installed_files
            if installed_file.path.endswith(".py")
        ]
        process = subprocess.Popen(
            args=args + py_files, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        _, stderr = process.communicate()
        if process.returncode != 0:
            pex_warnings.warn(
                "Failed to compile some .py files for install of {wheel} to {dest}:\n"
                "{stderr}".format(wheel=wheel_path, dest=dest, stderr=stderr.decode("utf-8"))
            )
        for root, _, files in os.walk(commonpath(py_files)):
            for f in files:
                if f.endswith(".pyc"):
                    file = InstalledFile(path=os.path.relpath(os.path.join(root, f), dest))
                    installed_files.append(file)

    dist = Distribution(location=dest, metadata=wheel.dist_metadata())
    entry_points = dist.get_entry_map()
    for entry_point in itertools.chain.from_iterable(
        entry_points.get(key, {}).values() for key in ("console_scripts", "gui_scripts")
    ):
        script_abspath = os.path.join(install_paths.scripts, entry_point.name)
        with safe_open(script_abspath, "w") as fp:
            fp.write(
                dedent(
                    """\
                    {shebang}
                    # -*- coding: utf-8 -*-
                    import importlib
                    import sys

                    object_ref = "{object_ref}"
                    modname, qualname_separator, qualname = object_ref.partition(':')
                    entry_point = importlib.import_module(modname)
                    if qualname_separator:
                        for attr in qualname.split('.'):
                            entry_point = getattr(entry_point, attr)

                    if __name__ == '__main__':
                        sys.exit(entry_point())
                    """
                ).format(
                    shebang=interpreter.shebang() if interpreter else "#!python",
                    object_ref=str(entry_point),
                )
            )
        chmod_plus_x(fp.name)
        installed_files.append(
            InstalledWheel.create_installed_file(path=script_abspath, dest_dir=dest)
        )

    with safe_open(os.path.join(dest, wheel.metadata_path("INSTALLER")), "w") as fp:
        print("pex", file=fp)
    installed_files.append(InstalledWheel.create_installed_file(path=fp.name, dest_dir=dest))

    if interpreter:
        # Finalize a proper venv install with REQUESTED and a RECORD to support un-installing.
        if requested:
            requested_path = os.path.join(dest, wheel.metadata_path("REQUESTED"))
            touch(requested_path)
            installed_files.append(
                InstalledWheel.create_installed_file(path=requested_path, dest_dir=dest)
            )

        installed_files.append(InstalledFile(path=record_relpath, hash=None, size=None))
        Record.write(dst=record_abspath, installed_files=installed_files)

    return wheel



# Copyright 2021 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import sys

import pytest

from testing import run_pex_command


@pytest.mark.skipif(
    sys.version_info[:2] >= (3, 12),
    reason=(
        "There is an indirect urllib3 dependency which embeds six which uses a meta path importer "
        "that only implements the PEP-302 finder spec and not the modern spec. Only the modern "
        "finder spec is supported by Python 3.12+."
    ),
)
def test_pip_2020_resolver_engaged():
    # type: () -> None

    # The Pip legacy resolver cannot solve the following requirements but the 2020 resolver can.
    # Use this fact to prove we're plumbing Pip resolver version arguments correctly.
    pex_args = ["boto3==1.15.6", "botocore>1.17,<1.20", "--", "-c", "import boto3"]

    results = run_pex_command(
        args=["--resolver-version", "pip-legacy-resolver"] + pex_args, quiet=True
    )
    results.assert_failure()
    assert "Failed to resolve compatible distributions:" in results.error
    assert (
        "1: boto3==1.15.6 requires botocore<1.19.0,>=1.18.6 but 1 incompatible dist was resolved: "
        "botocore-1.19.63-py2.py3-none-any.whl" in results.error
    ), results.error

    run_pex_command(args=["--resolver-version", "pip-2020-resolver"] + pex_args).assert_success()



# Copyright 2021 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
import sys
from textwrap import dedent

import pytest

from pex.typing import TYPE_CHECKING
from testing import built_wheel, run_pex_command, run_simple_pex

if TYPE_CHECKING:
    from typing import Any


@pytest.mark.skipif(
    sys.version_info[:2] >= (3, 12),
    reason="We need to use setuptools<66 but Python 3.12+ require greater.",
)
def test_resolve_arbitrary_equality(tmpdir):
    # type: (Any) -> None

    def prepare_project(project_dir):
        # type: (str) -> None
        with open(os.path.join(project_dir, "pyproject.toml"), "w") as fp:
            fp.write(
                dedent(
                    """\
                    [build-system]
                    # Setuptools 66 removed support for PEP-440 non-compliant versions.
                    # See: https://setuptools.pypa.io/en/stable/history.html#v66-0-0
                    requires = ["setuptools<66"]
                    """
                )
            )

    with built_wheel(
        prepare_project=prepare_project,
        name="foo",
        version="1.0.2-fba4511",
        # We need this to allow the invalid version above to sneak by pip wheel metadata
        # verification.
        verify=False,
        python_requires=">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*",
    ) as whl:
        pex_root = os.path.join(str(tmpdir), "pex_root")
        pex_file = os.path.join(str(tmpdir), "pex")
        results = run_pex_command(
            args=["-o", pex_file, "--pex-root", pex_root, "--runtime-pex-root", pex_root, whl]
        )
        results.assert_success()

        output, returncode = run_simple_pex(pex_file, args=["-c", "import foo"])
        assert returncode == 0, output
        assert output == b""



import os
import subprocess
import sys

import pytest

from pex.typing import TYPE_CHECKING
from testing import run_pex_command

if TYPE_CHECKING:
    from typing import Any


@pytest.mark.skipif(sys.version_info < (3, 8), reason="Flask 2.3.3 requires Python3.8+.")
def test_standard_library_is_included(
    tmpdir,  # type: Any
):
    # type: (...) -> None

    built_pex_path = os.path.join(str(tmpdir), "flask.pex")
    run_pex_command(
        args=[
            "Flask==2.3.3",
            "-o",
            built_pex_path,
            "--venv",
        ]
    ).assert_success()

    output_path = os.path.join(str(tmpdir), "output.txt")
    script = """
import sys
error = ""

try:
    sys.path.append("{}")
    import __pex__
    import flask
except Exception as e:
    error = str(e)

with open("{}", 'w') as f:
    f.write(error)
""".format(
        built_pex_path, output_path
    )

    script_path = os.path.join(str(tmpdir), "script.py")
    with open(script_path, "w") as f:
        f.write(script)

    subprocess.check_call([sys.executable, script_path])

    with open(output_path, "r") as f:
        output = f.read()

    if output:
        pytest.fail("Test failed with exception: {}".format(output))



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import, print_function

import os
import sys

import pex_build
from hatchling.metadata.plugin.interface import MetadataHookInterface

if pex_build.TYPE_CHECKING:
    from typing import Any, Dict


def expand_value(
    value,  # type: Any
    **fmt  # type: str
):
    # type: (...) -> Any
    if isinstance(value, str):
        return value.format(**fmt)
    if isinstance(value, list):
        return [expand_value(val) for val in value]
    if isinstance(value, dict):
        return {key: expand_value(value, **fmt) for key, value in value.items()}
    return value


class AdjustMetadata(MetadataHookInterface):
    """Allows modifying project metadata.

    The following mutations are supported:
    + Specifying alternate requires-python metadata via _PEX_REQUIRES_PYTHON env var.
    + Expanding format string placeholder (`{name}`) with metadata values via the `expand` mapping
      of placeholder name to metadata value.
    """

    PLUGIN_NAME = "pex-adjust-metadata"

    def update(self, metadata):
        # type: (Dict[str, Any]) -> None

        dynamic_metadata = self.config.get("project")
        if not dynamic_metadata:
            return

        self._update_requires_python(metadata, dynamic_metadata)
        self._expand_placeholders(metadata, dynamic_metadata)

    def _update_requires_python(
        self,
        metadata,  # type: Dict[str, Any]
        dynamic_metadata,  # type: Dict[str, Any]
    ):
        # type: (...) -> None

        dynamic_requires_python = os.environ.get("_PEX_REQUIRES_PYTHON")
        static_requires_python = dynamic_metadata.get("requires-python")

        if dynamic_requires_python:
            if not static_requires_python:
                raise ValueError(
                    "A dynamic override of requires-python metadata was specified via "
                    "`_PEX_REQUIRES_PYTHON={dynamic_requires_python}` but there was no "
                    "corresponding static value defined in "
                    "`tool.hatch.metadata.hooks.{plugin_name}.project`.".format(
                        plugin_name=self.PLUGIN_NAME,
                        dynamic_requires_python=dynamic_requires_python,
                    )
                )

            print(
                "pex_build: Dynamically modifying pyproject.toml requires-python of {original} to "
                "{dynamic}".format(
                    original=static_requires_python, dynamic=dynamic_requires_python
                ),
                file=sys.stderr,
            )
            dynamic_metadata["requires-python"] = dynamic_requires_python

        if dynamic_metadata["requires-python"]:
            metadata["requires-python"] = dynamic_metadata["requires-python"]

    def _expand_placeholders(
        self,
        metadata,  # type: Dict[str, Any]
        dynamic_metadata,  # type: Dict[str, Any]
    ):
        # type: (...) -> None

        expand = self.config.get("expand")
        if not expand:
            return

        metadata.update(
            (
                key,
                expand_value(value, **{key: metadata[value] for key, value in expand.items()}),
            )
            for key, value in dynamic_metadata.items()
        )



#!/usr/bin/env python3

import argparse
import subprocess
import sys


def run_autoflake(*args: str) -> None:
    excludes = [
        # The `pex.compatibility` module contains funky conditional symbol export code based on the
        # Python version that trips up autoflake / pyflakes; so we omit the file in its entirety
        # instead of littering it with 10s of `# noqa`.
        "pex/compatibility.py",
        "pex/vendor/_vendored",
    ]
    subprocess.run(
        args=[
            "autoflake",
            *args,
            "--remove-all-unused-imports",
            "--remove-unused-variables",
            "--remove-rhs-for-unused-variables",
            "--in-place",
            "--exclude",
            ",".join(excludes),
            "--recursive",
            "build-backend",
            "docs",
            "pex",
            "scripts",
            "testing",
            "tests",
        ],
        check=True,
    )


def main(check: bool = False) -> None:
    if check:
        run_autoflake("--check-diff", "--quiet")
    else:
        run_autoflake()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--check", action="store_true")
    options = parser.parse_args()
    try:
        main(check=options.check)
    except subprocess.CalledProcessError as e:
        sys.exit(e.returncode)



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
import re
import subprocess
import sys

import pytest

from pex.layout import Layout
from pex.typing import TYPE_CHECKING
from testing import run_pex_command

if TYPE_CHECKING:
    from typing import Any


@pytest.mark.parametrize(
    "reqs_pex_layout", [pytest.param(layout, id=str(layout)) for layout in Layout.values()]
)
@pytest.mark.parametrize(
    "out_pex_layout", [pytest.param(layout, id=str(layout)) for layout in Layout.values()]
)
def test_requirements_pex_wheel_type_mismatch(
    tmpdir,  # type: Any
    reqs_pex_layout,  # type: Layout.Value
    out_pex_layout,  # type: Layout.Value
):
    # type: (...) -> None

    pre_installed_reqs_pex = os.path.join(str(tmpdir), "pre_installed_reqs.pex")
    run_pex_command(
        args=["cowsay==5.0", "--layout", str(reqs_pex_layout), "-o", pre_installed_reqs_pex]
    ).assert_success()

    wheel_file_reqs_pex = os.path.join(str(tmpdir), "wheel_file_reqs.pex")
    run_pex_command(
        args=[
            "cowsay==5.0",
            "--no-pre-install-wheels",
            "--layout",
            str(reqs_pex_layout),
            "-o",
            wheel_file_reqs_pex,
        ]
    ).assert_success()

    pex = os.path.join(str(tmpdir), "pex")

    def assert_pex():
        # type: () -> None
        assert (
            "5.0"
            == subprocess.check_output(args=[sys.executable, pex, "--version"])
            .decode("utf-8")
            .strip()
        )

    run_pex_command(
        args=[
            "--requirements-pex",
            pre_installed_reqs_pex,
            "-c" "cowsay",
            "--layout",
            str(out_pex_layout),
            "-o",
            pex,
        ]
    ).assert_success()
    assert_pex()

    run_pex_command(
        args=[
            "--requirements-pex",
            wheel_file_reqs_pex,
            "--no-pre-install-wheels",
            "-c",
            "cowsay",
            "--layout",
            str(out_pex_layout),
            "-o",
            pex,
        ]
    ).assert_success()
    assert_pex()

    # N.B.: We cannot currently re-materialize wheel files from pre-installed wheel chroots.
    # See: https://github.com/pex-tool/pex/issues/2299
    run_pex_command(
        args=["--no-pre-install-wheels", "--requirements-pex", pre_installed_reqs_pex], quiet=True
    ).assert_failure(
        expected_error_re=re.escape(
            "Cannot resolve .whl files from PEX at {reqs_pex}; its dependencies are in the form of "
            "pre-installed wheel chroots.".format(reqs_pex=pre_installed_reqs_pex)
        )
    )

    # But we can go the other way around and turn wheel files into pre-installed wheel chroots.
    run_pex_command(args=["--requirements-pex", wheel_file_reqs_pex], quiet=True).assert_success()



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
from textwrap import dedent

from pex.compatibility import commonpath
from pex.targets import LocalInterpreter
from pex.typing import TYPE_CHECKING
from testing import run_pex_command
from testing.cli import run_pex3

if TYPE_CHECKING:
    from typing import Any, Iterable, Optional


def test_lock_use_no_build_wheel(tmpdir):
    # type: (Any)-> None

    lock = os.path.join(str(tmpdir), "lock")
    run_pex3(
        "lock",
        "create",
        "ansicolors==1.1.8",
        "--only-binary",
        "ansicolors",
        "-o",
        lock,
        "--style",
        "universal",
        "--indent",
        "2",
    ).assert_success()

    def assert_pex_from_lock(
        extra_args=(),  # type: Iterable[str]
        expected_error=None,  # type: Optional[str]
    ):
        # type: (...) -> None

        pex_root = os.path.realpath(os.path.join(str(tmpdir), "pex_root"))
        args = [
            "--pex-root",
            pex_root,
            "--runtime-pex-root",
            pex_root,
            "--lock",
            lock,
            "--",
            "-c",
            "import colors, os; print(os.path.realpath(colors.__file__))",
        ]
        result = run_pex_command(args=list(extra_args) + args)
        if expected_error:
            result.assert_failure()
            assert expected_error in result.error
        else:
            result.assert_success()
            assert pex_root == commonpath((pex_root, result.output.strip()))

    assert_pex_from_lock()

    # A redundant --only-binary should not matter.
    assert_pex_from_lock(extra_args=["--only-binary", "ansicolors"])

    # An extraneous --only-build should not matter.
    assert_pex_from_lock(extra_args=["--only-build", "cowsay"])

    # However; a conflicting --only-build should matter.
    assert_pex_from_lock(
        extra_args=["--only-build", "ansicolors"],
        expected_error=dedent(
            """\
            Failed to resolve all requirements for {target_description} from {lock}:

            Configured with:
                build: True
                only_build: ansicolors
                use_wheel: True

            Dependency on ansicolors not satisfied, 1 incompatible candidate found:
            1.) ansicolors 1.1.8 (via: ansicolors==1.1.8) does not have any compatible artifacts:
            """
        ).format(lock=lock, target_description=LocalInterpreter.create().render_description()),
    )



# Copyright 2023 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os.path
from textwrap import dedent

from pex.interpreter import PythonInterpreter
from pex.typing import TYPE_CHECKING
from testing import run_pex_command

if TYPE_CHECKING:
    from typing import Any


def test_derive_consistent_shebang_platforms(
    tmpdir,  # type: Any
    current_interpreter,  # type: PythonInterpreter
):
    # type: (...) -> None

    pex = os.path.join(str(tmpdir), "pex")

    def read_pex_shebang():
        # type: () -> bytes
        with open(pex, "rb") as fp:
            return fp.readline()

    run_pex_command(args=["--platform", "linux_x86_64-cp-311-cp311", "-o", pex]).assert_success()
    assert b"#!/usr/bin/env python3.11\n" == read_pex_shebang()

    run_pex_command(
        args=[
            "--platform",
            "linux_x86_64-cp-311-cp311",
            "--platform",
            "macosx_10.9_x86_64-cp-311-cp311",
            "-o",
            pex,
        ]
    ).assert_success()
    assert b"#!/usr/bin/env python3.11\n" == read_pex_shebang()

    run_pex_command(
        args=[
            "--platform",
            "linux_x86_64-cp-3.11.5-cp311",
            "--platform",
            "macosx_10.9_x86_64-cp-311-cp311",
            "-o",
            pex,
        ]
    ).assert_success()
    assert b"#!/usr/bin/env python3.11\n" == read_pex_shebang()

    result = run_pex_command(
        args=[
            "--platform",
            "linux_x86_64-cp-310-cp310",
            "--platform",
            "macosx_10.9_x86_64-cp-311-cp311",
            "-o",
            pex,
        ]
    )
    result.assert_success()
    current_interpreter_shebang = current_interpreter.identity.hashbang()
    assert (
        "{shebang}\n".format(shebang=current_interpreter_shebang).encode("utf-8")
        == read_pex_shebang()
    )
    assert (
        dedent(
            """\
            PEXWarning: Could not calculate a targeted shebang for:
            abbreviated platform cp310-cp310-linux_x86_64
            abbreviated platform cp311-cp311-macosx_10_9_x86_64

            Using shebang: {shebang}
            If this is not appropriate, you can specify a custom shebang using the --python-shebang option.
            """
        ).format(shebang=current_interpreter_shebang)
        in result.error
    )



# Copyright 2024 Pex project contributors.
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import os.path
from argparse import Namespace, _ActionsContainer

from pex.compatibility import urlparse
from pex.fetcher import URLFetcher
from pex.orderedset import OrderedSet
from pex.pep_440 import Version
from pex.scie import science
from pex.scie.model import (
    ScieConfiguration,
    ScieInfo,
    ScieOptions,
    SciePlatform,
    ScieStyle,
    ScieTarget,
)
from pex.scie.science import SCIENCE_RELEASES_URL, SCIENCE_REQUIREMENT
from pex.typing import TYPE_CHECKING, cast
from pex.variables import ENV, Variables

if TYPE_CHECKING:
    from typing import Iterator, Optional, Tuple, Union


__all__ = (
    "ScieConfiguration",
    "ScieInfo",
    "SciePlatform",
    "ScieStyle",
    "ScieTarget",
    "build",
    "extract_options",
    "register_options",
    "render_options",
)


def register_options(parser):
    # type: (_ActionsContainer) -> None

    parser.add_argument(
        "--scie",
        "--par",
        dest="scie_style",
        default=None,
        type=ScieStyle.for_value,
        choices=ScieStyle.values(),
        help=(
            "Create one or more native executable scies from your PEX that include a portable "
            "CPython interpreter along with your PEX making for a truly hermetic PEX that can run "
            "on machines with no Python installed at all. If your PEX has multiple targets, "
            "whether `--platform`s, `--complete-platform`s or local interpreters in any "
            "combination, then one PEX scie will be made for each platform, selecting the latest "
            "compatible portable CPython interpreter. Note that only CPython>=3.8 is supported. If "
            "you'd like to explicitly control the target platforms or the exact portable CPython "
            "selected, see `--scie-platform`, `--scie-pbs-release` and `--scie-python-version`. "
            "Specifying `--scie {lazy}` will fetch the portable CPython interpreter just in time "
            "on first boot of the PEX scie on a given machine if needed. The URL(s) to fetch the "
            "portable CPython interpreter from can be customized by exporting the "
            "PEX_BOOTSTRAP_URLS environment variable pointing to a json file with the format: "
            '`{{"ptex": {{<file name 1>: <url>, ...}}}}` where the file names should match those '
            "found via `SCIE=inspect <the PEX scie> | jq .ptex` with appropriate replacement URLs. "
            "Specifying `--scie {eager}` will embed the portable CPython interpreter in your PEX "
            "scie making for a larger file, but requiring no internet access to boot. If you have "
            "customization needs not addressed by the Pex `--scie*` options, consider using "
            "`science` to build your scies (which is what Pex uses behind the scenes); see: "
            "https://science.scie.app.".format(lazy=ScieStyle.LAZY, eager=ScieStyle.EAGER)
        ),
    )
    parser.add_argument(
        "--scie-platform",
        dest="scie_platforms",
        default=[],
        action="append",
        type=SciePlatform.parse,
        choices=[
            platform
            for platform in SciePlatform.values()
            if platform not in (SciePlatform.WINDOWS_AARCH64, SciePlatform.WINDOWS_X86_64)
        ],
        help=(
            "The platform to produce the native PEX scie executable for. Can be specified multiple "
            "times. You can use a value of 'current' to select the current platform. If left "
            "unspecified, the platforms implied by the targets selected to build the PEX with are "
            "used. Those targets are influenced by the current interpreter running Pex as well as "
            "use of `--python`, `--interpreter-constraint`, `--platform` or `--complete-platform` "
            "options."
        ),
    )
    parser.add_argument(
        "--scie-pbs-release",
        dest="scie_pbs_release",
        default=None,
        type=str,
        help=(
            "The Python Standalone Builds release to use. Currently releases are dates of the form "
            "YYYYMMDD, e.g.: '20240713'. See their GitHub releases page at "
            "https://github.com/indygreg/python-build-standalone/releases to discover available "
            "releases. If left unspecified the latest release is used. N.B.: The latest lookup is "
            "cached for 5 days. To force a fresh lookup you can remove the cache at "
            "<USER CACHE DIR>/science/downloads."
        ),
    )
    parser.add_argument(
        "--scie-python-version",
        dest="scie_python_version",
        default=None,
        type=Version,
        help=(
            "The portable CPython version to select. Can be either in `<major>.<minor>` form; "
            "e.g.: '3.11', or else fully specified as `<major>.<minor>.<patch>`; e.g.: '3.11.3'. "
            "If you don't specify this option, Pex will do its best to guess appropriate portable "
            "CPython versions. N.B.: Python Standalone Builds does not provide all patch versions; "
            "so you should check their releases at "
            "https://github.com/indygreg/python-build-standalone/releases if you wish to pin down "
            "to the patch level."
        ),
    )
    parser.add_argument(
        "--scie-science-binary",
        dest="scie_science_binary",
        default=None,
        type=str,
        help=(
            "The file path of a `science` binary or a URL to use to fetch the `science` binary "
            "when there is no `science` on the PATH with a version matching {science_requirement}. "
            "Pex uses the official `science` releases at {science_releases_url} by default.".format(
                science_requirement=SCIENCE_REQUIREMENT, science_releases_url=SCIENCE_RELEASES_URL
            )
        ),
    )


def render_options(options):
    # type: (ScieOptions) -> str

    args = ["--scie", str(options.style)]
    for platform in options.platforms:
        args.append("--scie-platform")
        args.append(str(platform))
    if options.pbs_release:
        args.append("--scie-pbs-release")
        args.append(options.pbs_release)
    if options.python_version:
        args.append("--scie-python-version")
        args.append(".".join(map(str, options.python_version)))
    if options.science_binary_url:
        args.append("--scie-science-binary")
        args.append(options.science_binary_url)
    return " ".join(args)


def extract_options(options):
    # type: (Namespace) -> Optional[ScieOptions]

    if not options.scie_style:
        return None

    python_version = None  # type: Optional[Union[Tuple[int, int], Tuple[int, int, int]]]
    if options.scie_python_version:
        if (
            not options.scie_python_version.parsed_version.release
            or len(options.scie_python_version.parsed_version.release) < 2
        ):
            raise ValueError(
                "Invalid Python version: '{python_version}'.\n"
                "Must be in the form `<major>.<minor>` or `<major>.<minor>.<patch>`".format(
                    python_version=options.scie_python_version
                )
            )
        python_version = cast(
            "Union[Tuple[int, int], Tuple[int, int, int]]",
            options.scie_python_version.parsed_version.release,
        )
        if python_version < (3, 8):
            raise ValueError(
                "Invalid Python version: '{python_version}'.\n"
                "Scies are built using Python Standalone Builds which only supports Python >=3.8.\n"
                "To find supported Python versions, you can browse the releases here:\n"
                "  https://github.com/indygreg/python-build-standalone/releases".format(
                    python_version=options.scie_python_version
                )
            )

    science_binary_url = options.scie_science_binary
    if science_binary_url:
        url_info = urlparse.urlparse(options.scie_science_binary)
        if not url_info.scheme and url_info.path and os.path.isfile(url_info.path):
            science_binary_url = "file://{path}".format(path=os.path.abspath(url_info.path))

    return ScieOptions(
        style=options.scie_style,
        platforms=tuple(OrderedSet(options.scie_platforms)),
        pbs_release=options.scie_pbs_release,
        python_version=python_version,
        science_binary_url=science_binary_url,
    )


def build(
    configuration,  # type: ScieConfiguration
    pex_file,  # type: str
    url_fetcher=None,  # type: Optional[URLFetcher]
    env=ENV,  # type: Variables
):
    # type: (...) -> Iterator[ScieInfo]

    return science.build(configuration, pex_file, url_fetcher=url_fetcher, env=env)
